import logging
import re
from typing import List, Optional, Set, cast

import datahub.emitter.mce_builder as builder
from datahub.configuration.common import ConfigModel
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.transformer.dataset_transformer import OwnershipTransformer
from datahub.metadata.schema_classes import (
    OwnerClass,
    OwnershipClass,
    OwnershipTypeClass,
)
from datahub.metadata.urns import CorpGroupUrn, CorpUserUrn
from datahub.utilities.urns._urn_base import Urn
from datahub.utilities.urns.error import InvalidUrnError

logger = logging.getLogger(__name__)


class PatternCleanUpOwnershipConfig(ConfigModel):
    pattern_for_cleanup: List[str]


class PatternCleanUpOwnership(OwnershipTransformer):
    """Transformer that clean the ownership URN."""

    ctx: PipelineContext
    config: PatternCleanUpOwnershipConfig

    def __init__(self, config: PatternCleanUpOwnershipConfig, ctx: PipelineContext):
        super().__init__()
        self.ctx = ctx
        self.config = config

    @classmethod
    def create(
        cls, config_dict: dict, ctx: PipelineContext
    ) -> "PatternCleanUpOwnership":
        config = PatternCleanUpOwnershipConfig.parse_obj(config_dict)
        return cls(config, ctx)

    def _get_current_owner_urns(self, entity_urn: str) -> Set[str]:
        if self.ctx.graph is not None:
            current_ownership = self.ctx.graph.get_ownership(entity_urn=entity_urn)
            if current_ownership is not None:
                current_owner_urns: Set[str] = {
                    owner.owner for owner in current_ownership.owners
                }
                return current_owner_urns
            else:
                return set()
        else:
            return set()

    def _process_owner(self, name: str) -> str:
        for value in self.config.pattern_for_cleanup:
            name = re.sub(value, "", name)
        return name

    def transform_aspect(
        self, entity_urn: str, aspect_name: str, aspect: Optional[builder.Aspect]
    ) -> Optional[builder.Aspect]:
        # get current owner URNs from the graph
        current_owner_urns = self._get_current_owner_urns(entity_urn)

        # clean all the owners based on the parameters received from config
        cleaned_owner_urns: List[str] = []
        for owner_urn in current_owner_urns:
            username = ""
            try:
                owner: Urn = Urn.from_string(owner_urn)
                if isinstance(owner, CorpUserUrn):
                    username = str(CorpUserUrn(self._process_owner(owner.username)))
                elif isinstance(owner, CorpGroupUrn):
                    username = str(CorpGroupUrn(self._process_owner(owner.name)))
                else:
                    logger.warning(f"{owner_urn} is not a supported owner type.")
                    username = owner_urn
            except InvalidUrnError:
                logger.warning(f"Could not parse {owner_urn} from {entity_urn}")
                username = owner_urn
            cleaned_owner_urns.append(username)

        ownership_type, ownership_type_urn = builder.validate_ownership_type(
            OwnershipTypeClass.TECHNICAL_OWNER
        )
        owners = [
            OwnerClass(
                owner=owner,
                type=ownership_type,
                typeUrn=ownership_type_urn,
            )
            for owner in cleaned_owner_urns
        ]

        out_ownership_aspect: OwnershipClass = OwnershipClass(
            owners=[],
            lastModified=None,
        )

        # generate the ownership aspect for the cleaned users
        out_ownership_aspect.owners.extend(owners)
        return cast(Optional[builder.Aspect], out_ownership_aspect)



from typing import Type

import pydantic

from datahub.ingestion.api.global_context import get_graph_context


def auto_connection_resolver(
    connection_field: str = "connection",
) -> classmethod:
    def _resolve_connection(cls: Type, values: dict) -> dict:
        if connection_field in values:
            connection_urn = values.pop(connection_field)

            graph = get_graph_context()
            if not graph:
                raise ValueError(
                    "Fetching connection details from the backend requires a DataHub graph client."
                )

            conn = graph.get_connection_json(connection_urn)
            if conn is None:
                raise ValueError(
                    f"Connection {connection_urn} not found using {graph}."
                )

            # TODO: Should this do some additional validation against the config model?

            # Update the config, but don't overwrite existing values.
            for key, value in conn.items():
                if key not in values:
                    values[key] = value

        return values

    # Hack: Pydantic maintains unique list of validators by referring its __name__.
    # https://github.com/pydantic/pydantic/blob/v1.10.9/pydantic/main.py#L264
    # This hack ensures that multiple validators do not overwrite each other.
    _resolve_connection.__name__ = f"{_resolve_connection.__name__}_{connection_field}"
    return pydantic.root_validator(pre=True, allow_reuse=True)(_resolve_connection)



import logging
import pathlib
import tempfile
from collections import OrderedDict
from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Dict, Iterable, List, Optional, Set, Tuple

import lkml
import lkml.simple
from looker_sdk.error import SDKError

from datahub.configuration.git import GitInfo
from datahub.emitter.mce_builder import make_schema_field_urn
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.emitter.mcp_builder import gen_containers
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import MetadataWorkUnitProcessor, SourceCapability
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.common.subtypes import (
    BIContainerSubTypes,
    DatasetSubTypes,
)
from datahub.ingestion.source.git.git_import import GitClone
from datahub.ingestion.source.looker.looker_common import (
    CORPUSER_DATAHUB,
    LookerExplore,
    LookerUtil,
    LookerViewId,
    ViewField,
    ViewFieldType,
    ViewFieldValue,
    deduplicate_fields,
    gen_project_key,
)
from datahub.ingestion.source.looker.looker_connection import (
    get_connection_def_based_on_connection_string,
)
from datahub.ingestion.source.looker.looker_dataclasses import LookerConstant
from datahub.ingestion.source.looker.looker_lib_wrapper import LookerAPI
from datahub.ingestion.source.looker.looker_template_language import (
    load_and_preprocess_file,
)
from datahub.ingestion.source.looker.looker_view_id_cache import (
    LookerModel,
    LookerViewFileLoader,
    LookerViewIdCache,
)
from datahub.ingestion.source.looker.lookml_concept_context import (
    LookerFieldContext,
    LookerViewContext,
)
from datahub.ingestion.source.looker.lookml_config import (
    BASE_PROJECT_NAME,
    MODEL_FILE_EXTENSION,
    VIEW_FILE_EXTENSION,
    LookerConnectionDefinition,
    LookMLSourceConfig,
    LookMLSourceReport,
)
from datahub.ingestion.source.looker.lookml_refinement import LookerRefinementResolver
from datahub.ingestion.source.looker.view_upstream import (
    AbstractViewUpstream,
    create_view_upstream,
)
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalHandler,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionSourceBase,
)
from datahub.metadata.com.linkedin.pegasus2avro.common import BrowsePaths, Status
from datahub.metadata.com.linkedin.pegasus2avro.dataset import (
    DatasetLineageTypeClass,
    FineGrainedLineageDownstreamType,
    UpstreamClass,
    UpstreamLineage,
    ViewProperties,
)
from datahub.metadata.com.linkedin.pegasus2avro.metadata.snapshot import DatasetSnapshot
from datahub.metadata.com.linkedin.pegasus2avro.mxe import MetadataChangeEvent
from datahub.metadata.schema_classes import (
    AuditStampClass,
    BrowsePathEntryClass,
    BrowsePathsV2Class,
    ContainerClass,
    DatasetPropertiesClass,
    FineGrainedLineageClass,
    FineGrainedLineageUpstreamTypeClass,
    SubTypesClass,
)
from datahub.sql_parsing.sqlglot_lineage import ColumnRef

VIEW_LANGUAGE_LOOKML: str = "lookml"
VIEW_LANGUAGE_SQL: str = "sql"

logger = logging.getLogger(__name__)


@dataclass
class LookerView:
    id: LookerViewId
    absolute_file_path: str
    connection: LookerConnectionDefinition
    upstream_dataset_urns: List[str]
    fields: List[ViewField]
    raw_file_content: str
    view_details: Optional[ViewProperties] = None

    @classmethod
    def determine_view_file_path(
        cls, base_folder_path: str, absolute_file_path: str
    ) -> str:
        splits: List[str] = absolute_file_path.split(base_folder_path, 1)
        if len(splits) != 2:
            logger.debug(
                f"base_folder_path({base_folder_path}) and absolute_file_path({absolute_file_path}) not matching"
            )
            return ViewFieldValue.NOT_AVAILABLE.value

        file_path: str = splits[1]
        logger.debug(f"file_path={file_path}")

        return file_path.strip(
            "/"
        )  # strip / from path to make it equivalent to source_file attribute of LookerModelExplore API

    @classmethod
    def from_looker_dict(
        cls,
        project_name: str,
        model_name: str,
        view_context: LookerViewContext,
        looker_view_id_cache: LookerViewIdCache,
        reporter: LookMLSourceReport,
        max_file_snippet_length: int,
        config: LookMLSourceConfig,
        ctx: PipelineContext,
        extract_col_level_lineage: bool = False,
        populate_sql_logic_in_descriptions: bool = False,
    ) -> Optional["LookerView"]:
        view_name = view_context.name()

        logger.debug(f"Handling view {view_name} in model {model_name}")

        looker_view_id: LookerViewId = LookerViewId(
            project_name=project_name,
            model_name=model_name,
            view_name=view_name,
            file_path=view_context.view_file_name(),
        )

        view_upstream: AbstractViewUpstream = create_view_upstream(
            view_context=view_context,
            looker_view_id_cache=looker_view_id_cache,
            config=config,
            ctx=ctx,
            reporter=reporter,
        )

        field_type_vs_raw_fields = OrderedDict(
            {
                ViewFieldType.DIMENSION: view_context.dimensions(),
                ViewFieldType.DIMENSION_GROUP: view_context.dimension_groups(),
                ViewFieldType.MEASURE: view_context.measures(),
            }
        )  # in order to maintain order in golden file

        view_fields: List[ViewField] = []

        for field_type, fields in field_type_vs_raw_fields.items():
            for field in fields:
                upstream_column_ref: List[ColumnRef] = []
                if extract_col_level_lineage:
                    upstream_column_ref = view_upstream.get_upstream_column_ref(
                        field_context=LookerFieldContext(raw_field=field)
                    )

                view_fields.append(
                    ViewField.view_fields_from_dict(
                        field_dict=field,
                        upstream_column_ref=upstream_column_ref,
                        type_cls=field_type,
                        populate_sql_logic_in_descriptions=populate_sql_logic_in_descriptions,
                    )
                )

        # special case where view is defined as derived sql, however fields are not defined
        if (
            len(view_fields) == 0
            and view_context.is_sql_based_derived_view_without_fields_case()
        ):
            view_fields = view_upstream.create_fields()

        view_fields = deduplicate_fields(view_fields)

        # Prep "default" values for the view, which will be overridden by the logic below.
        view_logic = view_context.view_file.raw_file_content[:max_file_snippet_length]

        if view_context.is_sql_based_derived_case():
            view_logic = view_context.sql()
            view_details = ViewProperties(
                materialized=False,
                viewLogic=view_logic,
                viewLanguage=VIEW_LANGUAGE_SQL,
            )
        elif view_context.is_native_derived_case():
            # We want this to render the full lkml block
            # e.g. explore_source: source_name { ... }
            # As such, we use the full derived_table instead of the explore_source.
            view_logic = str(lkml.dump(view_context.derived_table()))[
                :max_file_snippet_length
            ]
            view_lang = VIEW_LANGUAGE_LOOKML

            materialized = view_context.is_materialized_derived_view()

            view_details = ViewProperties(
                materialized=materialized, viewLogic=view_logic, viewLanguage=view_lang
            )
        else:
            view_details = ViewProperties(
                materialized=False,
                viewLogic=view_logic,
                viewLanguage=VIEW_LANGUAGE_LOOKML,
            )

        return LookerView(
            id=looker_view_id,
            absolute_file_path=view_context.view_file.absolute_file_path,
            connection=view_context.view_connection,
            upstream_dataset_urns=view_upstream.get_upstream_dataset_urn(),
            fields=view_fields,
            raw_file_content=view_context.view_file.raw_file_content,
            view_details=view_details,
        )


@dataclass
class LookerRemoteDependency:
    name: str
    url: str
    ref: Optional[str]


@dataclass
class LookerManifest:
    # This must be set if the manifest has local_dependency entries.
    # See https://cloud.google.com/looker/docs/reference/param-manifest-project-name
    project_name: Optional[str]
    constants: Optional[List[Dict[str, str]]]

    local_dependencies: List[str]
    remote_dependencies: List[LookerRemoteDependency]


@platform_name("Looker")
@config_class(LookMLSourceConfig)
@support_status(SupportStatus.CERTIFIED)
@capability(
    SourceCapability.PLATFORM_INSTANCE,
    "Use the `platform_instance` and `connection_to_platform_map` fields",
)
@capability(SourceCapability.LINEAGE_COARSE, "Supported by default")
@capability(
    SourceCapability.LINEAGE_FINE,
    "Enabled by default, configured using `extract_column_level_lineage`",
)
class LookMLSource(StatefulIngestionSourceBase):
    """
    This plugin extracts the following:
    - LookML views from model files in a project
    - Name, upstream table names, metadata for dimensions, measures, and dimension groups attached as tags
    - If API integration is enabled (recommended), resolves table and view names by calling the Looker API, otherwise supports offline resolution of these names.

    :::note
    To get complete Looker metadata integration (including Looker dashboards and charts and lineage to the underlying Looker views, you must ALSO use the `looker` source module.
    :::
    """

    platform = "lookml"

    def __init__(self, config: LookMLSourceConfig, ctx: PipelineContext):
        super().__init__(config, ctx)
        self.source_config: LookMLSourceConfig = config
        self.ctx = ctx
        self.reporter = LookMLSourceReport()

        # To keep track of projects (containers) which have already been ingested
        self.processed_projects: List[str] = []

        # This is populated during the git clone step.
        self.base_projects_folder: Dict[str, pathlib.Path] = {}
        self.remote_projects_git_info: Dict[str, GitInfo] = {}

        self.looker_client: Optional[LookerAPI] = None
        if self.source_config.api:
            self.looker_client = LookerAPI(self.source_config.api)
            self.reporter._looker_api = self.looker_client
            try:
                self.looker_client.all_connections()
            except SDKError as err:
                raise ValueError(
                    "Failed to retrieve connections from looker client. Please check to ensure that you have "
                    "manage_models permission enabled on this API key."
                ) from err

        self.manifest_constants: Dict[str, "LookerConstant"] = {}

    def _load_model(self, path: str) -> LookerModel:
        logger.debug(f"Loading model from file {path}")

        parsed = load_and_preprocess_file(
            path=path,
            reporter=self.reporter,
            source_config=self.source_config,
        )

        looker_model = LookerModel.from_looker_dict(
            parsed,
            BASE_PROJECT_NAME,
            self.source_config.project_name,
            self.base_projects_folder,
            path,
            self.source_config,
            self.reporter,
        )
        return looker_model

    def _get_upstream_lineage(
        self, looker_view: LookerView
    ) -> Optional[UpstreamLineage]:
        upstream_dataset_urns = looker_view.upstream_dataset_urns

        # Generate the upstream + fine grained lineage objects.
        upstreams = []
        observed_lineage_ts = datetime.now(tz=timezone.utc)
        for upstream_dataset_urn in upstream_dataset_urns:
            upstream = UpstreamClass(
                dataset=upstream_dataset_urn,
                type=DatasetLineageTypeClass.VIEW,
                auditStamp=AuditStampClass(
                    time=int(observed_lineage_ts.timestamp() * 1000),
                    actor=CORPUSER_DATAHUB,
                ),
            )
            upstreams.append(upstream)

        fine_grained_lineages: List[FineGrainedLineageClass] = []

        for field in looker_view.fields:
            fine_grained_lineages.append(
                FineGrainedLineageClass(
                    upstreamType=FineGrainedLineageUpstreamTypeClass.FIELD_SET,
                    upstreams=[
                        make_schema_field_urn(cll_ref.table, cll_ref.column)
                        for cll_ref in field.upstream_fields
                    ],
                    downstreamType=FineGrainedLineageDownstreamType.FIELD,
                    downstreams=[
                        make_schema_field_urn(
                            looker_view.id.get_urn(self.source_config),
                            field.name,
                        )
                    ],
                )
            )

        if upstreams:
            return UpstreamLineage(
                upstreams=upstreams, fineGrainedLineages=fine_grained_lineages or None
            )
        else:
            return None

    def _get_custom_properties(self, looker_view: LookerView) -> DatasetPropertiesClass:
        assert self.source_config.base_folder  # this is always filled out
        base_folder = self.base_projects_folder.get(
            looker_view.id.project_name, self.source_config.base_folder
        )
        try:
            file_path = str(
                pathlib.Path(looker_view.absolute_file_path).relative_to(
                    base_folder.resolve()
                )
            )
        except Exception:
            file_path = None
            logger.warning(
                f"Failed to resolve relative path for file {looker_view.absolute_file_path} w.r.t. folder {self.source_config.base_folder}"
            )

        custom_properties = {
            "looker.file.path": file_path or looker_view.absolute_file_path,
            "looker.model": looker_view.id.model_name,
        }
        dataset_props = DatasetPropertiesClass(
            name=looker_view.id.view_name, customProperties=custom_properties
        )

        maybe_git_info = self.source_config.project_dependencies.get(
            looker_view.id.project_name,
            self.remote_projects_git_info.get(looker_view.id.project_name),
        )
        if isinstance(maybe_git_info, GitInfo):
            git_info: Optional[GitInfo] = maybe_git_info
        else:
            git_info = self.source_config.git_info
        if git_info is not None and file_path:
            # It should be that looker_view.id.project_name is the base project.
            github_file_url = git_info.get_url_for_file_path(file_path)
            dataset_props.externalUrl = github_file_url

        return dataset_props

    def _build_dataset_mcps(
        self, looker_view: LookerView
    ) -> List[MetadataChangeProposalWrapper]:
        view_urn = looker_view.id.get_urn(self.source_config)

        subTypeEvent = MetadataChangeProposalWrapper(
            entityUrn=view_urn,
            aspect=SubTypesClass(typeNames=[DatasetSubTypes.VIEW]),
        )
        events = [subTypeEvent]
        if looker_view.view_details is not None:
            viewEvent = MetadataChangeProposalWrapper(
                entityUrn=view_urn,
                aspect=looker_view.view_details,
            )
            events.append(viewEvent)

        project_key = gen_project_key(self.source_config, looker_view.id.project_name)

        container = ContainerClass(container=project_key.as_urn())
        events.append(
            MetadataChangeProposalWrapper(entityUrn=view_urn, aspect=container)
        )

        events.append(
            MetadataChangeProposalWrapper(
                entityUrn=view_urn,
                aspect=looker_view.id.get_browse_path_v2(self.source_config),
            )
        )

        return events

    def _build_dataset_mce(self, looker_view: LookerView) -> MetadataChangeEvent:
        """
        Creates MetadataChangeEvent for the dataset, creating upstream lineage links
        """
        logger.debug(f"looker_view = {looker_view.id}")

        dataset_snapshot = DatasetSnapshot(
            urn=looker_view.id.get_urn(self.source_config),
            aspects=[],  # we append to this list later on
        )
        browse_paths = BrowsePaths(
            paths=[looker_view.id.get_browse_path(self.source_config)]
        )

        dataset_snapshot.aspects.append(browse_paths)
        dataset_snapshot.aspects.append(Status(removed=False))
        upstream_lineage = self._get_upstream_lineage(looker_view)
        if upstream_lineage is not None:
            dataset_snapshot.aspects.append(upstream_lineage)
        schema_metadata = LookerUtil._get_schema(
            self.source_config.platform_name,
            looker_view.id.view_name,
            looker_view.fields,
            self.reporter,
        )
        if schema_metadata is not None:
            dataset_snapshot.aspects.append(schema_metadata)
        dataset_snapshot.aspects.append(self._get_custom_properties(looker_view))

        return MetadataChangeEvent(proposedSnapshot=dataset_snapshot)

    def get_project_name(self, model_name: str) -> str:
        if self.source_config.project_name is not None:
            return self.source_config.project_name

        assert self.looker_client is not None, (
            "Failed to find a configured Looker API client"
        )
        try:
            model = self.looker_client.lookml_model(model_name, fields="project_name")
            assert model.project_name is not None, (
                f"Failed to find a project name for model {model_name}"
            )
            return model.project_name
        except SDKError:
            raise ValueError(
                f"Could not locate a project name for model {model_name}. Consider configuring a static project name "
                f"in your config file"
            ) from None

    def get_manifest_if_present(self, folder: pathlib.Path) -> Optional[LookerManifest]:
        manifest_file = folder / "manifest.lkml"

        if not manifest_file.exists():
            self.reporter.info(
                message="manifest.lkml file missing from project",
                context=str(manifest_file),
            )
            return None

        manifest_dict = load_and_preprocess_file(
            path=manifest_file,
            source_config=self.source_config,
            reporter=self.reporter,
        )

        manifest = LookerManifest(
            project_name=manifest_dict.get("project_name"),
            constants=manifest_dict.get("constants", []),
            local_dependencies=[
                x["project"] for x in manifest_dict.get("local_dependencys", [])
            ],
            remote_dependencies=[
                LookerRemoteDependency(name=x["name"], url=x["url"], ref=x.get("ref"))
                for x in manifest_dict.get("remote_dependencys", [])
            ],
        )
        return manifest

    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        return [
            *super().get_workunit_processors(),
            StaleEntityRemovalHandler.create(
                self, self.source_config, self.ctx
            ).workunit_processor,
        ]

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        with tempfile.TemporaryDirectory("lookml_tmp") as tmp_dir:
            # Clone the base_folder if necessary.
            if not self.source_config.base_folder:
                assert self.source_config.git_info
                # we don't have a base_folder, so we need to clone the repo and process it locally
                start_time = datetime.now()
                checkout_dir = self.source_config.git_info.clone(
                    tmp_path=tmp_dir,
                )
                self.reporter.git_clone_latency = datetime.now() - start_time
                self.source_config.base_folder = checkout_dir.resolve()

            self.base_projects_folder[BASE_PROJECT_NAME] = (
                self.source_config.base_folder
            )

            visited_projects: Set[str] = set()

            # We clone everything that we're pointed at.
            for project, p_ref in self.source_config.project_dependencies.items():
                # If we were given GitHub info, we need to clone the project.
                if isinstance(p_ref, GitInfo):
                    try:
                        p_checkout_dir = p_ref.clone(
                            tmp_path=f"{tmp_dir}/_included_/{project}",
                            # If a deploy key was provided, use it. Otherwise, fall back
                            # to the main project deploy key, if present.
                            fallback_deploy_key=(
                                self.source_config.git_info.deploy_key
                                if self.source_config.git_info
                                else None
                            ),
                        )

                        p_ref = p_checkout_dir.resolve()
                    except Exception as e:
                        logger.warning(
                            f"Failed to clone project dependency {project}. This can lead to failures in parsing lookml files later on: {e}",
                        )
                        visited_projects.add(project)
                        continue

                self.base_projects_folder[project] = p_ref

            self._recursively_check_manifests(
                tmp_dir,
                BASE_PROJECT_NAME,
                visited_projects,
                self.manifest_constants,
            )

            yield from self.get_internal_workunits()

            if not self.report.events_produced and not self.report.failures:
                # Don't pass if we didn't produce any events.
                self.report.report_failure(
                    "No Metadata Produced",
                    "No metadata was produced. Check the logs for more details.",
                )

    def _recursively_check_manifests(
        self,
        tmp_dir: str,
        project_name: str,
        project_visited: Set[str],
        manifest_constants: Dict[str, "LookerConstant"],
    ) -> None:
        if project_name in project_visited:
            return
        project_visited.add(project_name)

        project_path = self.base_projects_folder.get(project_name)
        if not project_path:
            logger.warning(
                f"Could not find {project_name} in the project_dependencies config. This can lead to failures in parsing lookml files later on.",
            )
            return

        manifest = self.get_manifest_if_present(project_path)
        if not manifest:
            return

        if manifest.constants:
            for constant in manifest.constants:
                if constant.get("name") and constant.get("value"):
                    manifest_constants[constant["name"]] = LookerConstant(
                        name=constant["name"],
                        value=constant["value"],
                    )

        # Special case handling if the root project has a name in the manifest file.
        if project_name == BASE_PROJECT_NAME and manifest.project_name:
            if (
                self.source_config.project_name is not None
                and manifest.project_name != self.source_config.project_name
            ):
                logger.warning(
                    f"The project name in the manifest file '{manifest.project_name}'"
                    f"does not match the configured project name '{self.source_config.project_name}'. "
                    "This can lead to failures in LookML include resolution and lineage generation."
                )
            elif self.source_config.project_name is None:
                self.source_config.project_name = manifest.project_name

        # Clone the remote project dependencies.
        for remote_project in manifest.remote_dependencies:
            if remote_project.name in project_visited:
                continue
            if remote_project.name in self.base_projects_folder:
                # In case a remote_dependency is specified in the project_dependencies config,
                # we don't need to clone it again.
                continue

            p_cloner = GitClone(f"{tmp_dir}/_remote_/{remote_project.name}")
            try:
                # TODO: For 100% correctness, we should be consulting
                # the manifest lock file for the exact ref to use.

                p_checkout_dir = p_cloner.clone(
                    ssh_key=(
                        self.source_config.git_info.deploy_key
                        if self.source_config.git_info
                        else None
                    ),
                    repo_url=remote_project.url,
                )

                self.base_projects_folder[remote_project.name] = (
                    p_checkout_dir.resolve()
                )
                repo = p_cloner.get_last_repo_cloned()
                assert repo
                remote_git_info = GitInfo(
                    url_template=remote_project.url,
                    repo="dummy/dummy",  # set to dummy values to bypass validation
                    branch=repo.active_branch.name,
                )
                remote_git_info.repo = (
                    ""  # set to empty because url already contains the full path
                )
                self.remote_projects_git_info[remote_project.name] = remote_git_info

            except Exception as e:
                logger.warning(
                    f"Failed to clone remote project {project_name}. This can lead to failures in parsing lookml files later on: {e}",
                )
                project_visited.add(project_name)
            else:
                self._recursively_check_manifests(
                    tmp_dir,
                    remote_project.name,
                    project_visited,
                    manifest_constants,
                )

        for project in manifest.local_dependencies:
            self._recursively_check_manifests(
                tmp_dir, project, project_visited, manifest_constants
            )

    def get_internal_workunits(self) -> Iterable[MetadataWorkUnit]:  # noqa: C901
        assert self.source_config.base_folder
        viewfile_loader = LookerViewFileLoader(
            self.source_config.project_name,
            self.base_projects_folder,
            self.reporter,
            self.source_config,
            self.manifest_constants,
        )
        logger.debug(f"LookML Constants : {', '.join(self.manifest_constants.keys())}")

        # Some views can be mentioned by multiple 'include' statements and can be included via different connections.

        # This map is used to keep track of which views files have already been processed
        # for a connection in order to prevent creating duplicate events.
        # Key: connection name, Value: view file paths
        processed_view_map: Dict[str, Set[str]] = {}

        # This map is used to keep track of the connection that a view is processed with.
        # Key: view unique identifier - determined by variables present in config `view_naming_pattern`
        # Value: Tuple(model file name, connection name)
        view_connection_map: Dict[str, Tuple[str, str]] = {}

        # The ** means "this directory and all subdirectories", and hence should
        # include all the files we want.
        model_files = sorted(
            self.source_config.base_folder.glob(f"**/*{MODEL_FILE_EXTENSION}")
        )
        model_suffix_len = len(".model")

        for file_path in model_files:
            self.reporter.report_models_scanned()
            model_name = file_path.stem[:-model_suffix_len]

            if not self.source_config.model_pattern.allowed(model_name):
                self.reporter.report_models_dropped(model_name)
                continue
            try:
                logger.debug(f"Attempting to load model: {file_path}")
                model = self._load_model(str(file_path))
            except Exception as e:
                self.reporter.report_warning(
                    title="Error Loading Model File",
                    message="Unable to load Looker model from file.",
                    context=f"Model Name: {model_name}, File Path: {file_path}",
                    exc=e,
                )
                continue

            assert model.connection is not None
            connection_definition = get_connection_def_based_on_connection_string(
                connection=model.connection,
                looker_client=self.looker_client,
                source_config=self.source_config,
                reporter=self.reporter,
            )

            if connection_definition is None:
                self.reporter.report_warning(
                    title="Failed to Load Connection",
                    message="Failed to load connection. Check your API key permissions and/or connection_to_platform_map configuration.",
                    context=f"Connection: {model.connection}",
                )
                self.reporter.report_models_dropped(model_name)
                continue

            explore_reachable_views: Set[str] = set()
            looker_refinement_resolver: LookerRefinementResolver = (
                LookerRefinementResolver(
                    looker_model=model,
                    connection_definition=connection_definition,
                    looker_viewfile_loader=viewfile_loader,
                    source_config=self.source_config,
                    reporter=self.reporter,
                )
            )

            if self.source_config.emit_reachable_views_only:
                model_explores_map = {d["name"]: d for d in model.explores}
                for explore_dict in model.explores:
                    try:
                        if LookerRefinementResolver.is_refinement(explore_dict["name"]):
                            continue

                        explore_dict = (
                            looker_refinement_resolver.apply_explore_refinement(
                                explore_dict
                            )
                        )
                        explore: LookerExplore = LookerExplore.from_dict(
                            model_name,
                            explore_dict,
                            model.resolved_includes,
                            viewfile_loader,
                            self.reporter,
                            model_explores_map,
                        )
                        if explore.upstream_views:
                            for view_name in explore.upstream_views:
                                explore_reachable_views.add(view_name.include)
                    except Exception as e:
                        self.reporter.report_warning(
                            title="Failed to process explores",
                            message="Failed to process explore dictionary.",
                            context=f"Explore Details: {explore_dict}",
                            exc=e,
                        )
                        logger.debug("Failed to process explore", exc_info=e)

            processed_view_files = processed_view_map.setdefault(
                model.connection, set()
            )

            project_name = self.get_project_name(model_name)

            looker_view_id_cache: LookerViewIdCache = LookerViewIdCache(
                project_name=project_name,
                model_name=model_name,
                looker_model=model,
                looker_viewfile_loader=viewfile_loader,
                reporter=self.reporter,
            )

            logger.debug(f"Model: {model_name}; Includes: {model.resolved_includes}")

            for include in model.resolved_includes:
                logger.debug(f"Considering {include} for model {model_name}")
                if include.include in processed_view_files:
                    logger.debug(f"view '{include}' already processed, skipping it")
                    continue
                logger.debug(f"Attempting to load view file: {include}")
                looker_viewfile = viewfile_loader.load_viewfile(
                    path=include.include,
                    project_name=include.project,
                    connection=connection_definition,
                    reporter=self.reporter,
                )

                if looker_viewfile is not None:
                    for raw_view in looker_viewfile.views:
                        raw_view_name = raw_view["name"]
                        if LookerRefinementResolver.is_refinement(raw_view_name):
                            continue

                        if (
                            self.source_config.emit_reachable_views_only
                            and raw_view_name not in explore_reachable_views
                        ):
                            logger.debug(
                                f"view {raw_view_name} is not reachable from an explore, skipping.."
                            )
                            self.reporter.report_unreachable_view_dropped(raw_view_name)
                            continue

                        self.reporter.report_views_scanned()
                        try:
                            raw_view = looker_refinement_resolver.apply_view_refinement(
                                raw_view=raw_view,
                            )

                            current_project_name: str = (
                                include.project
                                if include.project != BASE_PROJECT_NAME
                                else project_name
                            )

                            # if project is base project then it is available as self.base_projects_folder[
                            # _BASE_PROJECT_NAME]
                            base_folder_path: str = str(
                                self.base_projects_folder.get(
                                    current_project_name,
                                    self.base_projects_folder[BASE_PROJECT_NAME],
                                )
                            )

                            view_context: LookerViewContext = LookerViewContext(
                                raw_view=raw_view,
                                view_file=looker_viewfile,
                                view_connection=connection_definition,
                                view_file_loader=viewfile_loader,
                                looker_refinement_resolver=looker_refinement_resolver,
                                base_folder_path=base_folder_path,
                                reporter=self.reporter,
                            )

                            maybe_looker_view = LookerView.from_looker_dict(
                                project_name=current_project_name,
                                model_name=model_name,
                                view_context=view_context,
                                looker_view_id_cache=looker_view_id_cache,
                                reporter=self.reporter,
                                max_file_snippet_length=self.source_config.max_file_snippet_length,
                                extract_col_level_lineage=self.source_config.extract_column_level_lineage,
                                populate_sql_logic_in_descriptions=self.source_config.populate_sql_logic_for_missing_descriptions,
                                config=self.source_config,
                                ctx=self.ctx,
                            )
                        except Exception as e:
                            self.reporter.report_warning(
                                title="Error Loading View",
                                message="Unable to load Looker View.",
                                context=f"View Details: {raw_view}",
                                exc=e,
                            )

                            logger.debug(e, exc_info=e)

                            continue

                        if maybe_looker_view:
                            if self.source_config.view_pattern.allowed(
                                maybe_looker_view.id.view_name
                            ):
                                view_urn = maybe_looker_view.id.get_urn(
                                    self.source_config
                                )

                                view_connection_mapping = view_connection_map.get(
                                    view_urn
                                )
                                if not view_connection_mapping:
                                    view_connection_map[view_urn] = (
                                        model_name,
                                        model.connection,
                                    )
                                    # first time we are discovering this view
                                    logger.debug(
                                        f"Generating MCP for view {raw_view['name']}"
                                    )

                                    if (
                                        maybe_looker_view.id.project_name
                                        not in self.processed_projects
                                    ):
                                        yield from self.gen_project_workunits(
                                            maybe_looker_view.id.project_name
                                        )

                                        self.processed_projects.append(
                                            maybe_looker_view.id.project_name
                                        )

                                    for mcp in self._build_dataset_mcps(
                                        maybe_looker_view
                                    ):
                                        yield mcp.as_workunit()
                                    mce = self._build_dataset_mce(maybe_looker_view)
                                    yield MetadataWorkUnit(
                                        id=f"lookml-view-{maybe_looker_view.id}",
                                        mce=mce,
                                    )
                                    processed_view_files.add(include.include)
                                else:
                                    (
                                        prev_model_name,
                                        prev_model_connection,
                                    ) = view_connection_mapping
                                    if prev_model_connection != model.connection:
                                        # this view has previously been discovered and emitted using a different
                                        # connection
                                        logger.warning(
                                            f"view {maybe_looker_view.id.view_name} from model {model_name}, connection {model.connection} was previously processed via model {prev_model_name}, connection {prev_model_connection} and will likely lead to incorrect lineage to the underlying tables"
                                        )
                                        if not self.source_config.emit_reachable_views_only:
                                            logger.warning(
                                                "Consider enabling the `emit_reachable_views_only` flag to handle this case."
                                            )
                            else:
                                self.reporter.report_views_dropped(
                                    str(maybe_looker_view.id)
                                )

        if not self.source_config.emit_reachable_views_only:
            self.report_skipped_unreachable_views(viewfile_loader, processed_view_map)

        if (
            self.source_config.tag_measures_and_dimensions
            and self.reporter.events_produced != 0
        ):
            # Emit tag MCEs for measures and dimensions:
            for tag_mce in LookerUtil.get_tag_mces():
                yield MetadataWorkUnit(
                    id=f"tag-{tag_mce.proposedSnapshot.urn}", mce=tag_mce
                )

    def gen_project_workunits(self, project_name: str) -> Iterable[MetadataWorkUnit]:
        project_key = gen_project_key(
            self.source_config,
            project_name,
        )
        yield from gen_containers(
            container_key=project_key,
            name=project_name,
            sub_types=[BIContainerSubTypes.LOOKML_PROJECT],
        )
        yield MetadataChangeProposalWrapper(
            entityUrn=project_key.as_urn(),
            aspect=BrowsePathsV2Class(
                path=[BrowsePathEntryClass("Folders")],
            ),
        ).as_workunit()

    def report_skipped_unreachable_views(
        self,
        viewfile_loader: LookerViewFileLoader,
        processed_view_map: Optional[Dict[str, Set[str]]] = None,
    ) -> None:
        processed_view_map = processed_view_map or {}
        view_files: Dict[str, List[pathlib.Path]] = {}
        for project, folder_path in self.base_projects_folder.items():
            folder = pathlib.Path(folder_path)
            view_files[project] = list(folder.glob(f"**/*{VIEW_FILE_EXTENSION}"))

        skipped_view_paths: Dict[str, List[str]] = {}
        for project, views in view_files.items():
            skipped_paths: Set[str] = set()

            for view_path in views:
                # Check if the view is already in processed_view_map
                if not any(
                    str(view_path) in view_set
                    for view_set in processed_view_map.values()
                ):
                    looker_viewfile = viewfile_loader.load_viewfile(
                        path=str(view_path),
                        project_name=project,
                        connection=None,
                        reporter=self.reporter,
                    )

                    if looker_viewfile is not None:
                        for raw_view in looker_viewfile.views:
                            raw_view_name = raw_view.get("name", "")

                            if (
                                raw_view_name
                                and self.source_config.view_pattern.allowed(
                                    raw_view_name
                                )
                            ):
                                skipped_paths.add(str(view_path))

            skipped_view_paths[project] = list(skipped_paths)

        for project, view_paths in skipped_view_paths.items():
            for path in view_paths:
                self.reporter.report_warning(
                    title="Skipped View File",
                    message=(
                        "The Looker view file was skipped because it may not be referenced by any models."
                    ),
                    context=(f"Project: {project}, View File Path: {path}"),
                )

    def get_report(self):
        return self.reporter



from typing import Any, Iterable

import requests
import smart_open

from datahub.ingestion.fs.fs_base import FileInfo, FileSystem


class HttpFileSystem(FileSystem):
    @classmethod
    def create(cls, **kwargs):
        return HttpFileSystem()

    def open(self, path: str, **kwargs: Any) -> Any:
        return smart_open.open(path, mode="rb", transport_params=kwargs)

    def file_status(self, path: str) -> FileInfo:
        head = requests.head(path)
        if head.ok:
            return FileInfo(path, int(head.headers["Content-length"]), is_file=True)
        elif head.status_code == 404:
            raise FileNotFoundError(f"Requested path {path} does not exist.")
        else:
            raise IOError(f"Cannot get file status for the requested path {path}.")

    def list(self, path: str) -> Iterable[FileInfo]:
        status = self.file_status(path)
        return [status]



# mypy: ignore-errors
# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off
# isort: skip_file
from .....schema_classes import IncidentAssigneeClass
from .....schema_classes import IncidentInfoClass
from .....schema_classes import IncidentSourceClass
from .....schema_classes import IncidentSourceTypeClass
from .....schema_classes import IncidentStageClass
from .....schema_classes import IncidentStateClass
from .....schema_classes import IncidentStatusClass
from .....schema_classes import IncidentTypeClass


IncidentAssignee = IncidentAssigneeClass
IncidentInfo = IncidentInfoClass
IncidentSource = IncidentSourceClass
IncidentSourceType = IncidentSourceTypeClass
IncidentStage = IncidentStageClass
IncidentState = IncidentStateClass
IncidentStatus = IncidentStatusClass
IncidentType = IncidentTypeClass

# fmt: on



from functools import singledispatchmethod
from typing import List, Optional

from datahub.api.entities.assertion.assertion_operator import (
    BetweenOperator,
    ContainsOperator,
    EndsWithOperator,
    EqualToOperator,
    GreaterThanOperator,
    GreaterThanOrEqualToOperator,
    InOperator,
    IsFalseOperator,
    IsNullOperator,
    IsTrueOperator,
    LessThanOperator,
    LessThanOrEqualToOperator,
    MatchesRegexOperator,
    NotEqualToOperator,
    NotInOperator,
    NotNullOperator,
    Operators,
    StartsWithOperator,
)
from datahub.api.entities.assertion.field_assertion import (
    FieldTransform,
    FieldValuesAssertion,
)
from datahub.integrations.assertion.common import get_entity_name


class SnowflakeFieldValuesMetricSQLGenerator:
    @singledispatchmethod
    def values_metric_sql(
        self,
        operators: Operators,
        entity_name: str,
        transformed_field: str,
        where_clause: str,
    ) -> str:
        """
        Generates SQL that would return boolean value for each table row.
        1 if FAIL and 0 if PASS. Note the unusual reversal of 1 and 0.
        This is deliberate, as metric represents number of failing rows.
        """
        raise ValueError(f"Unsupported values metric operator type {type(operators)} ")

    @values_metric_sql.register
    def _(
        self,
        operators: InOperator,
        entity_name: str,
        transformed_field: str,
        where_clause: str,
    ) -> str:
        return f"""select case when {transformed_field} in {tuple(operators.value)} then 0 else 1 end
        from {entity_name} {where_clause}"""

    @values_metric_sql.register
    def _(
        self,
        operators: NotInOperator,
        entity_name: str,
        transformed_field: str,
        where_clause: str,
    ) -> str:
        return f"""select case when {transformed_field} not in {tuple(operators.value)} then 0 else 1 end
        from {entity_name} {where_clause}"""

    @values_metric_sql.register
    def _(
        self,
        operators: EqualToOperator,
        entity_name: str,
        transformed_field: str,
        where_clause: str,
    ) -> str:
        return f"""select case when {transformed_field} = {operators.value} then 0 else 1 end
        from {entity_name} {where_clause}"""

    @values_metric_sql.register
    def _(
        self,
        operators: NotEqualToOperator,
        entity_name: str,
        transformed_field: str,
        where_clause: str,
    ) -> str:
        return f"""select case when {transformed_field} != {operators.value} then 0 else 1 end
        from {entity_name} {where_clause}"""

    @values_metric_sql.register
    def _(
        self,
        operators: BetweenOperator,
        entity_name: str,
        transformed_field: str,
        where_clause: str,
    ) -> str:
        return f"""select case when {transformed_field} between {operators.min} and {operators.max} then 0 else 1 end
        from {entity_name} {where_clause}"""

    @values_metric_sql.register
    def _(
        self,
        operators: LessThanOperator,
        entity_name: str,
        transformed_field: str,
        where_clause: str,
    ) -> str:
        return f"""select case when {transformed_field} < {operators.value} then 0 else 1 end
        from {entity_name} {where_clause}"""

    @values_metric_sql.register
    def _(
        self,
        operators: LessThanOrEqualToOperator,
        entity_name: str,
        transformed_field: str,
        where_clause: str,
    ) -> str:
        return f"""select case when {transformed_field} <= {operators.value} then 0 else 1 end
        from {entity_name} {where_clause}"""

    @values_metric_sql.register
    def _(
        self,
        operators: GreaterThanOperator,
        entity_name: str,
        transformed_field: str,
        where_clause: str,
    ) -> str:
        return f"""select case when {transformed_field} > {operators.value} then 0 else 1 end
        from {entity_name} {where_clause}"""

    @values_metric_sql.register
    def _(
        self,
        operators: GreaterThanOrEqualToOperator,
        entity_name: str,
        transformed_field: str,
        where_clause: str,
    ) -> str:
        return f"""select case when {transformed_field} >= {operators.value} then 0 else 1 end
        from {entity_name} {where_clause}"""

    @values_metric_sql.register
    def _(
        self,
        operators: IsNullOperator,
        entity_name: str,
        transformed_field: str,
        where_clause: str,
    ) -> str:
        return f"""select case when {transformed_field} is null then 0 else 1 end
        from {entity_name} {where_clause}"""

    @values_metric_sql.register
    def _(
        self,
        operators: NotNullOperator,
        entity_name: str,
        transformed_field: str,
        where_clause: str,
    ) -> str:
        return f"""select case when {transformed_field} is not null then 0 else 1 end
        from {entity_name} {where_clause}"""

    @values_metric_sql.register
    def _(
        self,
        operators: IsTrueOperator,
        entity_name: str,
        transformed_field: str,
        where_clause: str,
    ) -> str:
        return f"""select case when {transformed_field} then 0 else 1 end
        from {entity_name} {where_clause}"""

    @values_metric_sql.register
    def _(
        self,
        operators: IsFalseOperator,
        entity_name: str,
        transformed_field: str,
        where_clause: str,
    ) -> str:
        return f"""select case when not {transformed_field} then 0 else 1 end
        from {entity_name} {where_clause}"""

    @values_metric_sql.register
    def _(
        self,
        operators: ContainsOperator,
        entity_name: str,
        transformed_field: str,
        where_clause: str,
    ) -> str:
        return f"""select case when contains({transformed_field},'{operators.value}') then 0 else 1 end
        from {entity_name} {where_clause}"""

    @values_metric_sql.register
    def _(
        self,
        operators: StartsWithOperator,
        entity_name: str,
        transformed_field: str,
        where_clause: str,
    ) -> str:
        return f"""select case when startswith({transformed_field},'{operators.value}') then 0 else 1 end
        from {entity_name} {where_clause}"""

    @values_metric_sql.register
    def _(
        self,
        operators: EndsWithOperator,
        entity_name: str,
        transformed_field: str,
        where_clause: str,
    ) -> str:
        return f"""select case when endswith({transformed_field},'{operators.value}') then 0 else 1 end
        from {entity_name} {where_clause}"""

    @values_metric_sql.register
    def _(
        self,
        operators: MatchesRegexOperator,
        entity_name: str,
        transformed_field: str,
        where_clause: str,
    ) -> str:
        return f"""select case when REGEXP_LIKE({transformed_field},'{operators.value}') then 0 else 1 end
        from {entity_name} {where_clause}"""

    def _setup_where_clause(self, filters: List[Optional[str]]) -> str:
        where_clause = " and ".join(f for f in filters if f)
        return f"where {where_clause}" if where_clause else ""

    def _setup_field_transform(
        self, field: str, transform: Optional[FieldTransform]
    ) -> str:
        if transform is None:
            return field
        elif transform is FieldTransform.LENGTH:
            return f"length({field})"
        raise ValueError(f"Unsupported transform type {transform}")

    def metric_sql(self, assertion: FieldValuesAssertion) -> str:
        """
        Note that this applies negative operator in order to check whether or not
        number of invalid value rows are less than configured failThreshold.

        Args:
            assertion (FieldValuesAssertion): _description_

        Returns:
            str: _description_
        """
        entity_name = ".".join(get_entity_name(assertion))

        dataset_filter = (
            assertion.filters.sql
            if assertion.filters and assertion.filters.sql
            else None
        )
        where_clause = self._setup_where_clause(
            [
                dataset_filter,
                f"{assertion.field} is not null" if assertion.exclude_nulls else None,
            ]
        )
        transformed_field = self._setup_field_transform(
            assertion.field, assertion.field_transform
        )
        # this sql would return boolean value for each table row.  1 if fail and 0 if pass.
        sql = self.values_metric_sql(
            assertion.operator, entity_name, transformed_field, where_clause
        )

        # metric would be number of failing rows OR percentage of failing rows.
        if assertion.failure_threshold.type == "count":
            return f"select sum($1) as metric from ({sql})"
        else:  # percentage
            return f"select sum($1)/count(*) as metric from ({sql})"



import json
import logging
import sys
from datetime import datetime
from typing import Any, Dict, Iterator, List, Optional, Tuple, Union
from urllib.parse import urlparse

import boto3
import botocore
import click

import datahub.metadata.schema_classes
from datahub.cli.cli_utils import post_entity
from datahub.configuration.common import GraphError
from datahub.ingestion.graph.client import DataHubGraph, get_default_graph
from datahub.metadata.schema_classes import SystemMetadataClass
from datahub.telemetry import telemetry

logger = logging.getLogger(__name__)

DEFAULT_CREDS_EXPIRY_DURATION_SECONDS = 60 * 60
DEFAULT_FABRIC_TYPE = datahub.metadata.schema_classes.FabricTypeClass.PROD

DATA_PLATFORM_INSTANCE_WAREHOUSE_ASPECT = "icebergWarehouseInfo"


@click.group()
def iceberg() -> None:
    """A group of commands to manage Iceberg warehouses using DataHub as the Iceberg Catalog."""
    pass


def validate_creds(client_id: str, client_secret: str, region: str) -> Any:
    try:
        # Create a boto3 client with the provided credentials
        # Using STS (Security Token Service) for validation
        sts_client = boto3.client(
            "sts",
            aws_access_key_id=client_id,
            aws_secret_access_key=client_secret,
            region_name=region,
        )

        # Try to get caller identity
        sts_client.get_caller_identity()

        # If successful, return True and the account info
        return sts_client

    except (
        botocore.exceptions.ClientError,
        botocore.exceptions.NoCredentialsError,
    ):
        # If credentials are invalid, return False with error message
        click.secho(
            "Invalid credentials",
            fg="red",
            err=True,
        )
        sys.exit(1)


def validate_role(role: str, sts_client: Any, duration_seconds: Optional[int]) -> None:
    try:
        session_name = (
            f"datahub-cli-iceberg-validation-{datetime.now().strftime('%Y%m%d%H%M%S')}"
        )
        # Assume the IAM role to ensure the settings we have are valid and if not, can report them at config time.

        # If duration_seconds is not specified, datahub will attempt to default to an internal default
        # defined in S3CredentialProvider.java DEFAULT_CREDS_DURATION_SECS. However, it is not possible to know for sure
        # if that value is permitted based on how the role is configured. So, during the configuration of the warehouse
        # we must attempt to use the intended expiration duration (default or explicitly supplied) to ensure it
        # actually does work.
        if duration_seconds is None:
            duration_seconds = DEFAULT_CREDS_EXPIRY_DURATION_SECONDS

        assumed_role = sts_client.assume_role(
            RoleArn=role,
            RoleSessionName=session_name,
            DurationSeconds=duration_seconds,
        )

        # Extract the temporary credentials
        credentials = assumed_role["Credentials"]
        return credentials

    except Exception as e:
        click.secho(
            f"Failed to assume role using '{role}' with error: {e}",
            fg="red",
            err=True,
        )
        sys.exit(1)


def validate_warehouse(data_root: str) -> None:
    # validate data_root location
    scheme = urlparse(data_root).scheme
    if scheme != "s3":
        click.secho(
            f"Unsupported warehouse location '{data_root}', supported schemes: s3",
            fg="red",
            err=True,
        )
        sys.exit(1)


@iceberg.command()
@click.option(
    "-w", "--warehouse", required=True, type=str, help="The name of the warehouse"
)
@click.option(
    "-p", "--description", required=False, type=str, help="Description of the warehouse"
)
@click.option(
    "-d",
    "--data_root",
    required=True,
    type=str,
    help="The path to the data root for the warehouse data",
)
@click.option(
    "-i",
    "--client_id",
    required=True,
    type=str,
    help="Client ID to authenticate with the storage provider of the data root",
)
@click.option(
    "-s",
    "--client_secret",
    required=True,
    type=str,
    help="Client Secret to authenticate with the storage provider of the data root",
)
@click.option(
    "-g",
    "--region",
    required=True,
    type=str,
    help="Storage provider specific region where the warehouse data root is located",
)
@click.option(
    "-r",
    "--role",
    required=True,
    type=str,
    help="Storage provider specific role to be used when vending credentials",
)
@click.option(
    "-e",
    "--env",
    required=False,
    type=str,
    help=f"Environment where all assets stored in this warehouse belong to. Defaults to {DEFAULT_FABRIC_TYPE} if unspecified",
)
@click.option(
    "-x",
    "--duration_seconds",
    required=False,
    type=int,
    help=f"Expiration duration for temporary credentials used for role. Defaults to {DEFAULT_CREDS_EXPIRY_DURATION_SECONDS} seconds if unspecified",
)
@telemetry.with_telemetry(capture_kwargs=["duration_seconds"])
def create(
    warehouse: str,
    description: Optional[str],
    data_root: str,
    client_id: str,
    client_secret: str,
    region: str,
    role: str,
    duration_seconds: Optional[int],
    env: Optional[str],
) -> None:
    """
    Create an iceberg warehouse.
    """

    client = get_default_graph()

    urn = iceberg_data_platform_instance_urn(warehouse)

    if client.exists(urn):
        click.secho(
            f"Warehouse with name {warehouse} already exists",
            fg="red",
            err=True,
        )
        sys.exit(1)

    # will throw an actionable error message if invalid.
    validate_warehouse(data_root)
    storage_client = validate_creds(client_id, client_secret, region)
    validate_role(role, storage_client, duration_seconds)

    client_id_urn, client_secret_urn = create_iceberg_secrets(
        client, warehouse, client_id, client_secret
    )

    if env is None:
        env = DEFAULT_FABRIC_TYPE

    warehouse_aspect = DATA_PLATFORM_INSTANCE_WAREHOUSE_ASPECT
    warehouse_aspect_obj: Dict[str, Any] = {
        "dataRoot": data_root,
        "clientId": client_id_urn,
        "clientSecret": client_secret_urn,
        "region": region,
        "role": role,
        "env": env,
    }

    if duration_seconds:
        warehouse_aspect_obj["tempCredentialExpirationSeconds"] = duration_seconds

    data_platform_instance_properties_aspect_obj = {
        "name": warehouse,
    }

    if description:
        data_platform_instance_properties_aspect_obj["description"] = description

    data_platform_instance_properties_aspect = "dataPlatformInstanceProperties"

    entity_type = "dataPlatformInstance"
    system_metadata: Union[None, SystemMetadataClass] = None

    post_entity(
        client._session,
        client.config.server,
        urn=urn,
        aspect_name=data_platform_instance_properties_aspect,
        entity_type=entity_type,
        aspect_value=data_platform_instance_properties_aspect_obj,
        system_metadata=system_metadata,
    )

    # If status is non 200, post_entity will raise an exception.

    post_entity(
        client._session,
        client.config.server,
        urn=urn,
        aspect_name=warehouse_aspect,
        entity_type=entity_type,
        aspect_value=warehouse_aspect_obj,
        system_metadata=system_metadata,
    )

    click.secho(
        f"✅ Created warehouse with urn {urn}, clientID: {client_id_urn}, and clientSecret: {client_secret_urn}",
        fg="green",
    )


@iceberg.command()
@click.option(
    "-w", "--warehouse", required=True, type=str, help="The name of the warehouse"
)
@click.option(
    "-p",
    "--description",
    required=False,
    type=str,
    help="Description of the warehouse",
)
@click.option(
    "-d",
    "--data_root",
    required=True,
    type=str,
    help="The path to the data root for the warehouse data",
)
@click.option(
    "-i",
    "--client_id",
    required=True,
    type=str,
    help="Client ID to authenticate with the storage provider of the data root",
)
@click.option(
    "-s",
    "--client_secret",
    required=True,
    type=str,
    help="Client Secret to authenticate with the storage provider of the data root",
)
@click.option(
    "-g",
    "--region",
    required=True,
    type=str,
    help="Storage provider specific region where the warehouse data root is located",
)
@click.option(
    "-r",
    "--role",
    required=True,
    type=str,
    help="Storage provider specific role to be used when vending credentials",
)
@click.option(
    "-e",
    "--env",
    required=False,
    type=str,
    help=f"Environment where all assets stored in this warehouse belong to. Defaults to {DEFAULT_FABRIC_TYPE} if unspecified",
)
@click.option(
    "-x",
    "--duration_seconds",
    required=False,
    type=int,
    help=f"Expiration duration for temporary credentials used for role. Defaults to {DEFAULT_CREDS_EXPIRY_DURATION_SECONDS} seconds if unspecified",
)
@telemetry.with_telemetry(capture_kwargs=["duration_seconds"])
def update(
    warehouse: str,
    data_root: str,
    description: Optional[str],
    client_id: str,
    client_secret: str,
    region: str,
    role: str,
    env: Optional[str],
    duration_seconds: Optional[int],
) -> None:
    """
    Update iceberg warehouses. Can only update credentials, and role. Cannot update region
    """

    client = get_default_graph()

    urn = iceberg_data_platform_instance_urn(warehouse)

    if not client.exists(urn):
        raise click.ClickException(f"Warehouse with name {warehouse} does not exist")

    validate_warehouse(data_root)
    storage_client = validate_creds(client_id, client_secret, region)
    validate_role(role, storage_client, duration_seconds)

    client_id_urn, client_secret_urn = update_iceberg_secrets(
        client, warehouse, client_id, client_secret
    )

    if env is None:
        env = DEFAULT_FABRIC_TYPE

    warehouse_aspect = DATA_PLATFORM_INSTANCE_WAREHOUSE_ASPECT
    warehouse_aspect_obj: Dict[str, Any] = {
        "dataRoot": data_root,
        "clientId": client_id_urn,
        "clientSecret": client_secret_urn,
        "region": region,
        "role": role,
        "env": env,
    }
    if duration_seconds:
        warehouse_aspect_obj["tempCredentialExpirationSeconds"] = duration_seconds

    data_platform_instance_properties_aspect_obj = {
        "name": warehouse,
    }

    if description:
        data_platform_instance_properties_aspect_obj["description"] = description

    data_platform_instance_properties_aspect = "dataPlatformInstanceProperties"

    entity_type = "dataPlatformInstance"
    system_metadata: Union[None, SystemMetadataClass] = None

    post_entity(
        client._session,
        client.config.server,
        urn=urn,
        aspect_name=data_platform_instance_properties_aspect,
        entity_type=entity_type,
        aspect_value=data_platform_instance_properties_aspect_obj,
        system_metadata=system_metadata,
    )

    # If status is non 200, post_entity will raise an exception.
    post_entity(
        client._session,
        client.config.server,
        urn=urn,
        aspect_name=warehouse_aspect,
        entity_type=entity_type,
        aspect_value=warehouse_aspect_obj,
        system_metadata=system_metadata,
    )

    click.secho(
        f"✅ Updated warehouse with urn {urn}, clientID: {client_id_urn}, and clientSecret: {client_secret_urn}",
        fg="green",
    )


@iceberg.command()
@telemetry.with_telemetry()
def list() -> None:
    """
    List iceberg warehouses
    """

    client = get_default_graph()

    for warehouse in get_all_warehouses(client):
        click.echo(warehouse)


@iceberg.command()
@click.option(
    "-w", "--warehouse", required=True, type=str, help="The name of the warehouse"
)
@telemetry.with_telemetry()
def get(warehouse: str) -> None:
    """Fetches the details of the specified iceberg warehouse"""
    client = get_default_graph()
    urn = iceberg_data_platform_instance_urn(warehouse)

    if client.exists(urn):
        warehouse_aspect = client.get_aspect(
            entity_urn=urn,
            aspect_type=datahub.metadata.schema_classes.IcebergWarehouseInfoClass,
        )
        click.echo(urn)
        if warehouse_aspect:
            click.echo(json.dumps(warehouse_aspect.to_obj(), sort_keys=True, indent=2))
    else:
        raise click.ClickException(f"Iceberg warehouse {warehouse} does not exist")


@iceberg.command()
@click.option(
    "-w", "--warehouse", required=True, type=str, help="The name of the warehouse"
)
@click.option("-n", "--dry-run", required=False, is_flag=True)
@click.option(
    "-f",
    "--force",
    required=False,
    is_flag=True,
    help="force the delete if set without confirmation",
)
@telemetry.with_telemetry(capture_kwargs=["dry_run", "force"])
def delete(warehouse: str, dry_run: bool, force: bool) -> None:
    """
    Delete warehouse
    """

    urn = iceberg_data_platform_instance_urn(warehouse)

    client = get_default_graph()

    if not client.exists(urn):
        raise click.ClickException(f"urn {urn} not found")

    # Confirm this is a managed warehouse by checking for presence of IcebergWarehouse aspect
    aspect = client.get_aspect(
        entity_urn=urn,
        aspect_type=datahub.metadata.schema_classes.IcebergWarehouseInfoClass,
    )
    if aspect:
        warehouse_aspect: datahub.metadata.schema_classes.IcebergWarehouseInfoClass = (
            aspect
        )

        urns_to_delete: List = []
        resource_names_to_be_deleted: List = []
        for entity in get_related_entities_for_platform_instance(client, urn):
            # Do we really need this double-check?
            if "__typename" in entity and "urn" in entity:
                if entity["__typename"] in ["Container", "Dataset"]:
                    # add the Platform Resource URN to also be deleted for each dataset.
                    # This is not user visible, so no need to show a name to the user and include it in the count. Each
                    # instance corresponds to a dataset whose name is shown.
                    if entity["__typename"] == "Dataset":
                        resource_urn = platform_resource_urn(
                            entity["properties"]["qualifiedName"]
                        )
                        urns_to_delete.append(resource_urn)

                    urns_to_delete.append(entity["urn"])
                    resource_names_to_be_deleted.append(
                        entity.get("name", entity.get("urn"))
                    )

        if dry_run:
            click.echo(
                f"[Dry-run] Would delete warehouse {urn} and the following datasets and namespaces"
            )
            for resource in resource_names_to_be_deleted:
                click.echo(f"   {resource}")
        else:
            if not force:
                click.confirm(
                    f"This will delete {warehouse} warehouse, credentials, and {len(resource_names_to_be_deleted)} datasets and namespaces from DataHub. Do you want to continue?",
                    abort=True,
                )

            # Delete the resources in the warehouse first, so that in case it is interrupted, the warehouse itself is
            # still available to enumerate the resources in it that are not yet deleted.
            for urn_to_delete in urns_to_delete:
                client.hard_delete_entity(urn_to_delete)

            client.hard_delete_entity(urn)
            client.hard_delete_entity(warehouse_aspect.clientId)
            client.hard_delete_entity(warehouse_aspect.clientSecret)

            click.echo(
                f"✅ Successfully deleted iceberg warehouse {warehouse} and associated credentials, {len(resource_names_to_be_deleted)} datasets and namespaces"
            )


def iceberg_data_platform_instance_urn(warehouse: str) -> str:
    return f"urn:li:dataPlatformInstance:({iceberg_data_platform()},{warehouse})"


def platform_resource_urn(dataset_name: str) -> str:
    return f"urn:li:platformResource:iceberg.{dataset_name}"


def iceberg_data_platform() -> str:
    return "urn:li:dataPlatform:iceberg"


def iceberg_client_id_urn(warehouse):
    return f"urn:li:dataHubSecret:{warehouse}-client_id"


def iceberg_client_secret_urn(warehouse):
    return f"urn:li:dataHubSecret:{warehouse}-client_secret"


def create_iceberg_secrets(
    client: DataHubGraph, warehouse: str, client_id: str, client_secret: str
) -> Tuple[str, str]:
    graphql_query = """
        mutation createIcebergSecrets($clientIdName: String!, $clientId: String!, $clientSecretName: String!, $clientSecret: String!) {
            createClientId: createSecret(
                input: {name: $clientIdName, value: $clientId}
            )
            createClientSecret: createSecret(
                input: {name: $clientSecretName, value: $clientSecret}
            )
        }
    """
    variables = {
        "clientIdName": f"{warehouse}-client_id",
        "clientId": client_id,
        "clientSecretName": f"{warehouse}-client_secret",
        "clientSecret": client_secret,
    }
    try:
        response = client.execute_graphql(
            graphql_query, variables=variables, format_exception=False
        )
    except GraphError as graph_error:
        try:
            error = json.loads(str(graph_error).replace('"', '\\"').replace("'", '"'))
            click.secho(
                f"Failed to save Iceberg warehouse credentials :{error[0]['message']}",
                fg="red",
                err=True,
            )
        except Exception:
            click.secho(
                f"Failed to save Iceberg warehouse credentials :\n{graph_error}",
                fg="red",
                err=True,
            )
        sys.exit(1)

    if "createClientId" in response and "createClientSecret" in response:
        return response["createClientId"], response["createClientSecret"]

    click.secho(
        f"Internal error: Unexpected response saving credentials:\n{response}",
        fg="red",
        err=True,
    )
    sys.exit(1)


def update_iceberg_secrets(
    client: DataHubGraph, warehouse: str, client_id: str, client_secret: str
) -> Tuple[str, str]:
    graphql_query = """
        mutation updateIcebergSecrets($clientIdUrn: String!, $clientIdName: String!, $clientId: String!, $clientSecretUrn: String!, $clientSecretName: String!, $clientSecret: String!) {
            updateClientId: updateSecret(
                input: {urn: $clientIdUrn, name: $clientIdName, value: $clientId}
            )
            updateClientSecret: updateSecret(
                input: {urn: $clientSecretUrn, name: $clientSecretName, value: $clientSecret}
            )
        }
    """
    variables = {
        "clientIdUrn": iceberg_client_id_urn(warehouse),
        "clientIdName": f"{warehouse}-client_id",
        "clientId": client_id,
        "clientSecretUrn": iceberg_client_secret_urn(warehouse),
        "clientSecretName": f"{warehouse}-client_secret",
        "clientSecret": client_secret,
    }
    try:
        response = client.execute_graphql(
            graphql_query, variables=variables, format_exception=False
        )
    except GraphError as graph_error:
        try:
            error = json.loads(str(graph_error).replace('"', '\\"').replace("'", '"'))
            click.secho(
                f"Failed to save Iceberg warehouse credentials :{error[0]['message']}",
                fg="red",
                err=True,
            )
        except Exception:
            click.secho(
                f"Failed to save Iceberg warehouse credentials :\n{graph_error}",
                fg="red",
                err=True,
            )
        sys.exit(1)

    if "updateClientId" in response and "updateClientSecret" in response:
        return response["updateClientId"], response["updateClientSecret"]

    click.secho(
        f"Internal error: Unexpected response saving credentials:\n{response}",
        fg="red",
        err=True,
    )
    sys.exit(1)


def get_all_warehouses(client: DataHubGraph) -> Iterator[str]:
    start: int = 0
    total = None
    graph_query = """
        query getIcebergWarehouses($start: Int, $count: Int) {
          search(
            input: {type: DATA_PLATFORM_INSTANCE, query: "dataPlatform:iceberg", start: $start, count: $count}
          ) {
            start
            total
            searchResults {
              entity {
                urn
                ... on DataPlatformInstance {
                  instanceId
                }
              }
            }
          }
        }
        """
    count = 10
    variables = {"start": start, "count": count}
    while total is None or start < total:
        response = client.execute_graphql(
            graph_query, variables=variables, format_exception=True
        )
        if "search" in response and "total" in response["search"]:
            total = response["search"]["total"]
            search_results = response["search"].get("searchResults", [])
            for result in search_results:
                yield result["entity"]["instanceId"]
            start += count
            variables = {"start": start, "count": count}
            # if total is not None and
        else:
            break


def get_related_entities_for_platform_instance(
    client: DataHubGraph, data_platform_instance_urn: str
) -> Iterator[Dict]:
    start: int = 0
    total = None

    graph_query = """
        query getIcebergResources($platformInstanceUrn: String!, $start: Int!, $count: Int!) {
          searchAcrossEntities(
            input: {types: [DATASET, CONTAINER], query: "*", start: $start, count: $count, orFilters: [{and: [{field: "platformInstance", values: [$platformInstanceUrn]}]}]}
          ) {
            start
            total
            searchResults {
              entity {
                __typename
                urn
                ... on Dataset {
                  urn
                  name
                  properties{
                    qualifiedName
                  }
                }
              }
            }
          }
        }
    """
    count = 10
    variables = {
        "start": start,
        "count": count,
        "platformInstanceUrn": data_platform_instance_urn,
    }
    while total is None or start < total:
        response = client.execute_graphql(
            graph_query, variables=variables, format_exception=True
        )
        if (
            "searchAcrossEntities" in response
            and "total" in response["searchAcrossEntities"]
        ):
            total = response["searchAcrossEntities"]["total"]
            search_results = response["searchAcrossEntities"].get("searchResults", [])
            for result in search_results:
                yield result["entity"]
            start += count
            variables["start"] = start
        else:
            break



from typing import Dict, List, Optional, Union

from datahub.emitter.mcp_patch_builder import MetadataPatchProposal
from datahub.metadata.schema_classes import (
    KafkaAuditHeaderClass,
    PropertyValueClass,
    StructuredPropertyDefinitionClass as StructuredPropertyDefinition,
    SystemMetadataClass,
)
from datahub.utilities.urns.urn import Urn


# This patch builder is for structured property entities. Not for the aspect on assets.
class StructuredPropertyPatchBuilder(MetadataPatchProposal):
    def __init__(
        self,
        urn: str,
        system_metadata: Optional[SystemMetadataClass] = None,
        audit_header: Optional[KafkaAuditHeaderClass] = None,
    ) -> None:
        super().__init__(
            urn, system_metadata=system_metadata, audit_header=audit_header
        )

    # can only be used when creating a new structured property
    def set_qualified_name(
        self, qualified_name: str
    ) -> "StructuredPropertyPatchBuilder":
        self._add_patch(
            StructuredPropertyDefinition.ASPECT_NAME,
            "add",
            path=("qualifiedName",),
            value=qualified_name,
        )
        return self

    def set_display_name(
        self, display_name: Optional[str] = None
    ) -> "StructuredPropertyPatchBuilder":
        if display_name is not None:
            self._add_patch(
                StructuredPropertyDefinition.ASPECT_NAME,
                "add",
                path=("displayName",),
                value=display_name,
            )
        return self

    # can only be used when creating a new structured property
    def set_value_type(
        self, value_type: Union[str, Urn]
    ) -> "StructuredPropertyPatchBuilder":
        self._add_patch(
            StructuredPropertyDefinition.ASPECT_NAME,
            "add",
            path=("valueType",),
            value=value_type,
        )
        return self

    # can only be used when creating a new structured property
    def set_type_qualifier(
        self, type_qualifier: Optional[Dict[str, List[str]]] = None
    ) -> "StructuredPropertyPatchBuilder":
        if type_qualifier is not None:
            self._add_patch(
                StructuredPropertyDefinition.ASPECT_NAME,
                "add",
                path=("typeQualifier",),
                value=type_qualifier,
            )
        return self

    # can only be used when creating a new structured property
    def add_allowed_value(
        self, allowed_value: PropertyValueClass
    ) -> "StructuredPropertyPatchBuilder":
        self._add_patch(
            StructuredPropertyDefinition.ASPECT_NAME,
            "add",
            path=("allowedValues", str(allowed_value.get("value"))),
            value=allowed_value,
        )
        return self

    def set_cardinality(self, cardinality: str) -> "StructuredPropertyPatchBuilder":
        self._add_patch(
            StructuredPropertyDefinition.ASPECT_NAME,
            "add",
            path=("cardinality",),
            value=cardinality,
        )
        return self

    def add_entity_type(
        self, entity_type: Union[str, Urn]
    ) -> "StructuredPropertyPatchBuilder":
        self._add_patch(
            StructuredPropertyDefinition.ASPECT_NAME,
            "add",
            path=("entityTypes", str(entity_type)),
            value=entity_type,
        )
        return self

    def set_description(
        self, description: Optional[str] = None
    ) -> "StructuredPropertyPatchBuilder":
        if description is not None:
            self._add_patch(
                StructuredPropertyDefinition.ASPECT_NAME,
                "add",
                path=("description",),
                value=description,
            )
        return self

    def set_immutable(self, immutable: bool) -> "StructuredPropertyPatchBuilder":
        self._add_patch(
            StructuredPropertyDefinition.ASPECT_NAME,
            "add",
            path=("immutable",),
            value=immutable,
        )
        return self



import doctest
from types import ModuleType


def assert_doctest(module: ModuleType) -> None:
    result = doctest.testmod(
        module,
        raise_on_error=True,
        verbose=True,
    )
    if result.attempted == 0:
        raise ValueError(f"No doctests found in {module.__name__}")



"""Testing utilities for the datahub package.

These modules are included in the `datahub.testing` namespace, but aren't
part of the public API. They're placed here, rather than in the `tests`
directory, so they can be used by other packages that depend on `datahub`.
"""



from __future__ import annotations

from typing import TYPE_CHECKING, Optional, overload

from datahub.errors import ItemNotFoundError, MultipleItemsFoundError, SdkUsageError
from datahub.ingestion.graph.client import DataHubGraph
from datahub.metadata.urns import (
    CorpUserUrn,
    DomainUrn,
    GlossaryTermUrn,
)
from datahub.sdk.search_filters import Filter, FilterDsl as F

if TYPE_CHECKING:
    from datahub.sdk.main_client import DataHubClient


class ResolverClient:
    def __init__(self, client: DataHubClient):
        self._client = client

    # TODO: add caching to this method

    @property
    def _graph(self) -> DataHubGraph:
        return self._client._graph

    def domain(self, *, name: str) -> DomainUrn:
        urn_str = self._graph.get_domain_urn_by_name(name)
        if urn_str is None:
            raise ItemNotFoundError(f"Domain with name {name} not found")
        return DomainUrn.from_string(urn_str)

    @overload
    def user(self, *, name: str) -> CorpUserUrn: ...
    @overload
    def user(self, *, email: str) -> CorpUserUrn: ...
    def user(
        self, *, name: Optional[str] = None, email: Optional[str] = None
    ) -> CorpUserUrn:
        filter_explanation: str
        filter: Filter
        if name is not None:
            if email is not None:
                raise SdkUsageError("Cannot specify both name and email for auto_user")
            # We're filtering on both fullName and displayName. It's not clear
            # what the right behavior is here.
            filter_explanation = f"with name {name}"
            filter = F.or_(
                F.custom_filter("fullName", "EQUAL", [name]),
                F.custom_filter("displayName", "EQUAL", [name]),
            )
        elif email is not None:
            filter_explanation = f"with email {email}"
            filter = F.custom_filter("email", "EQUAL", [email])
        else:
            raise SdkUsageError("Must specify either name or email for auto_user")

        filter = F.and_(
            F.entity_type(CorpUserUrn.ENTITY_TYPE),
            filter,
        )
        users = list(self._client.search.get_urns(filter=filter))
        if len(users) == 0:
            # TODO: In auto methods, should we just create the user/domain/etc if it doesn't exist?
            raise ItemNotFoundError(f"User {filter_explanation} not found")
        elif len(users) > 1:
            raise MultipleItemsFoundError(
                f"Multiple users found {filter_explanation}: {users}"
            )
        else:
            return CorpUserUrn.from_string(users[0])

    def term(self, *, name: str) -> GlossaryTermUrn:
        # TODO: Add some limits on the graph fetch
        terms = list(
            self._client.search.get_urns(
                filter=F.and_(
                    F.entity_type(GlossaryTermUrn.ENTITY_TYPE),
                    F.custom_filter("name", "EQUAL", [name]),
                ),
            )
        )
        if len(terms) == 0:
            raise ItemNotFoundError(f"Term with name {name} not found")
        elif len(terms) > 1:
            raise SdkUsageError(f"Multiple terms found with name {name}: {terms}")
        else:
            return GlossaryTermUrn.from_string(terms[0])



from typing import Optional, Type, Union

import pydantic


def pydantic_multiline_string(field: str) -> classmethod:
    """If the field is present and contains an escaped newline, replace it with a real newline.

    This makes the assumption that the field value is never supposed to have a
    r"\n" in it, and instead should only have newline characters. This is generally
    a safe assumption for SSH keys and similar.

    The purpose of this helper is to make us more forgiving of small formatting issues
    in recipes, without sacrificing correctness across the board.
    """

    def _validate_field(
        cls: Type, v: Union[None, str, pydantic.SecretStr]
    ) -> Optional[str]:
        if v is not None:
            if isinstance(v, pydantic.SecretStr):
                v = v.get_secret_value()
            v = v.replace(r"\n", "\n")

        return v

    # Hack: Pydantic maintains unique list of validators by referring its __name__.
    # https://github.com/pydantic/pydantic/blob/v1.10.9/pydantic/main.py#L264
    # This hack ensures that multiple field deprecated do not overwrite each other.
    _validate_field.__name__ = f"{_validate_field.__name__}_{field}"
    return pydantic.validator(field, pre=True, allow_reuse=True)(_validate_field)



from dataclasses import dataclass
from typing import Optional, Union

from datahub.emitter.kafka_emitter import DatahubKafkaEmitter, KafkaEmitterConfig
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import RecordEnvelope, WorkUnit
from datahub.ingestion.api.sink import Sink, SinkReport, WriteCallback
from datahub.metadata.com.linkedin.pegasus2avro.mxe import (
    MetadataChangeEvent,
    MetadataChangeProposal,
)


class KafkaSinkConfig(KafkaEmitterConfig):
    # This extra layer of indirection exists in case we need to add extra
    # config options to the sink config.
    pass


@dataclass
class _KafkaCallback:
    reporter: SinkReport
    record_envelope: RecordEnvelope
    write_callback: WriteCallback

    def kafka_callback(self, err: Optional[Exception], msg: str) -> None:
        if err is not None:
            self.reporter.report_failure(err)
            self.write_callback.on_failure(
                self.record_envelope, err, {"error": err, "msg": msg}
            )
        else:
            self.reporter.report_record_written(self.record_envelope)
            self.write_callback.on_success(self.record_envelope, {"msg": msg})


class DatahubKafkaSink(Sink[KafkaSinkConfig, SinkReport]):
    emitter: DatahubKafkaEmitter

    def __post_init__(self):
        self.emitter = DatahubKafkaEmitter(self.config)

    def handle_work_unit_start(self, workunit: WorkUnit) -> None:
        pass

    def handle_work_unit_end(self, workunit: WorkUnit) -> None:
        self.emitter.flush()

    def write_record_async(
        self,
        record_envelope: RecordEnvelope[
            Union[
                MetadataChangeEvent,
                MetadataChangeProposal,
                MetadataChangeProposalWrapper,
            ]
        ],
        write_callback: WriteCallback,
    ) -> None:
        callback = _KafkaCallback(
            self.report, record_envelope, write_callback
        ).kafka_callback
        try:
            record = record_envelope.record
            self.emitter.emit(
                record,
                callback=callback,
            )
        except Exception as err:
            # In case we throw an exception while trying to emit the record,
            # catch it and report the failure. This might happen if the schema
            # registry is down or otherwise misconfigured, in which case we'd
            # fail when serializing the record.
            callback(err, f"Failed to write record: {err}")

    def close(self) -> None:
        self.emitter.flush()



import shlex
from typing import List, Optional, Union

import requests
from requests.auth import HTTPBasicAuth


def _format_header(name: str, value: Union[str, bytes]) -> str:
    if name == "Authorization":
        return f"{name!s}: <redacted>"
    return f"{name!s}: {value!s}"


def make_curl_command(
    session: requests.Session, method: str, url: str, payload: Optional[str] = None
) -> str:
    fragments: List[str] = ["curl", "-X", method]

    for header_name, header_value in session.headers.items():
        fragments.extend(["-H", _format_header(header_name, header_value)])

    if session.auth:
        if isinstance(session.auth, HTTPBasicAuth):
            fragments.extend(["-u", f"{session.auth.username}:<redacted>"])
        else:
            # For other auth types, they should be handled via headers
            fragments.extend(["-H", "<unknown auth type>"])

    if payload:
        fragments.extend(["--data", payload])

    fragments.append(url)
    return shlex.join(fragments)



import os
from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional

from lark import Tree

from datahub.ingestion.source.powerbi.config import DataPlatformPair
from datahub.sql_parsing.sqlglot_lineage import ColumnLineageInfo

TRACE_POWERBI_MQUERY_PARSER = os.getenv("DATAHUB_TRACE_POWERBI_MQUERY_PARSER", False)


@dataclass
class IdentifierAccessor:
    """
    statement
        public_order_date = Source{[Schema="public",Item="order_date"]}[Data]
    will be converted to IdentifierAccessor instance
    where:

        "Source" is identifier

        "[Schema="public",Item="order_date"]" is "items" in ItemSelector. Data of items varies as per DataSource

        "public_order_date" is in "next" of ItemSelector. The "next" will be None if this identifier is leaf i.e., table

    """

    identifier: str
    items: Dict[str, Any]
    next: Optional["IdentifierAccessor"]


@dataclass
class DataAccessFunctionDetail:
    arg_list: Tree
    data_access_function_name: str
    identifier_accessor: Optional[IdentifierAccessor]


@dataclass
class ReferencedTable:
    warehouse: str
    catalog: Optional[str]
    database: str
    schema: str
    table: str


@dataclass
class DataPlatformTable:
    data_platform_pair: DataPlatformPair
    urn: str


@dataclass
class Lineage:
    upstreams: List[DataPlatformTable]
    column_lineage: List[ColumnLineageInfo]

    @staticmethod
    def empty() -> "Lineage":
        return Lineage(upstreams=[], column_lineage=[])


class FunctionName(Enum):
    NATIVE_QUERY = "Value.NativeQuery"
    POSTGRESQL_DATA_ACCESS = "PostgreSQL.Database"
    ORACLE_DATA_ACCESS = "Oracle.Database"
    SNOWFLAKE_DATA_ACCESS = "Snowflake.Databases"
    MSSQL_DATA_ACCESS = "Sql.Database"
    DATABRICK_DATA_ACCESS = "Databricks.Catalogs"
    GOOGLE_BIGQUERY_DATA_ACCESS = "GoogleBigQuery.Database"
    AMAZON_REDSHIFT_DATA_ACCESS = "AmazonRedshift.Database"
    DATABRICK_MULTI_CLOUD_DATA_ACCESS = "DatabricksMultiCloud.Catalogs"



import sys
from typing import List, Optional

if sys.version_info < (3, 10):
    from importlib_metadata import EntryPoint, entry_points
else:
    from importlib.metadata import EntryPoint, entry_points


_CUSTOM_PACKAGE_GROUP_KEY = "datahub.custom_packages"

_MODELS_KEY = "models"
_URNS_KEY = "urns"


class CustomPackageException(Exception):
    pass


def _get_all_registered_custom_packages() -> List[EntryPoint]:
    return list(entry_points(group=_CUSTOM_PACKAGE_GROUP_KEY))


def _get_custom_entrypoint_for_name(name: str) -> Optional[EntryPoint]:
    entrypoints = [
        ep for ep in _get_all_registered_custom_packages() if ep.name == name
    ]

    if not entrypoints:
        return None

    if len(entrypoints) > 1:
        all_package_options = [
            entrypoint.dist.name for entrypoint in entrypoints if entrypoint.dist
        ]
        raise CustomPackageException(
            f"Multiple custom packages registered for {name}: cannot pick between {all_package_options}"
        )

    return entrypoints[0]


def _get_custom_package_for_name(name: str) -> Optional[str]:
    entrypoint = _get_custom_entrypoint_for_name(name)
    if entrypoint:
        return entrypoint.value
    return None


def model_version_name() -> str:
    custom_entrypoint = _get_custom_entrypoint_for_name(_MODELS_KEY)
    if not custom_entrypoint:
        return "bundled"

    if not custom_entrypoint.dist:
        return custom_entrypoint.value

    custom_model_package = custom_entrypoint.dist.name
    version = custom_entrypoint.dist.version

    try:
        package = __import__(custom_model_package)

        nice_version_name = getattr(package, "nice_version_name", None)
        if nice_version_name:
            version = nice_version_name()
    except ImportError:
        pass

    return f"{custom_model_package} ({version})"


def get_custom_models_package() -> Optional[str]:
    return _get_custom_package_for_name(_MODELS_KEY)


def get_custom_urns_package() -> Optional[str]:
    return _get_custom_package_for_name(_URNS_KEY)



import types

import datahub.metadata.schema_classes as models
from datahub.errors import SdkUsageError
from datahub.ingestion.graph.config import DatahubClientConfig
from datahub.metadata.urns import (
    ChartUrn,
    ContainerUrn,
    CorpGroupUrn,
    CorpUserUrn,
    DashboardUrn,
    DataPlatformInstanceUrn,
    DataPlatformUrn,
    DatasetUrn,
    DomainUrn,
    GlossaryTermUrn,
    SchemaFieldUrn,
    TagUrn,
)
from datahub.sdk.container import Container
from datahub.sdk.dataset import Dataset
from datahub.sdk.main_client import DataHubClient
from datahub.sdk.search_filters import Filter, FilterDsl

# We want to print out the warning if people do `from datahub.sdk import X`.
# But we don't want to print out warnings if they're doing a more direct
# import like `from datahub.sdk.container import Container`, since that's
# what our internal code does.
_vars = {}
for _name, _value in list(locals().items()):
    if not _name.startswith("_") and (
        _name == "models" or not isinstance(_value, types.ModuleType)
    ):
        _vars[_name] = _value
        del locals()[_name]


def __getattr__(name):
    import warnings

    from datahub.errors import ExperimentalWarning

    warnings.warn(
        "The new datahub SDK (e.g. datahub.sdk.*) is experimental. "
        "Our typical backwards-compatibility and stability guarantees do not apply to this code. "
        "When it's promoted to stable, the import path will change "
        "from `from datahub.sdk import ...` to `from datahub import ...`.",
        ExperimentalWarning,
        stacklevel=2,
    )
    return _vars[name]



from datahub.metadata.urns import Urn

__all__ = ["Urn", "guess_entity_type"]


def guess_entity_type(urn: str) -> str:
    assert urn.startswith("urn:li:"), "urns must start with urn:li:"
    return urn.split(":")[2]



import logging
from dataclasses import dataclass
from typing import Optional

import cachetools
from pydantic import BaseModel, ValidationError

from datahub.api.entities.platformresource.platform_resource import (
    PlatformResource,
    PlatformResourceKey,
)
from datahub.ingestion.graph.client import DataHubGraph
from datahub.metadata.urns import TagUrn

logger: logging.Logger = logging.getLogger(__name__)


@dataclass
class BigQueryLabel:
    key: str
    value: Optional[str]

    def primary_key(self) -> str:
        return f"{self.key}/{self.value}" if self.value else f"{self.key}"


class BigQueryLabelInfo(BaseModel):
    datahub_urn: str
    managed_by_datahub: bool
    key: str
    value: str


@dataclass()
class BigQueryLabelPlatformResource:
    datahub_urn: str
    project: Optional[str]
    managed_by_datahub: bool
    label: BigQueryLabel

    def platform_resource_key(self) -> PlatformResourceKey:
        return PlatformResourceKey(
            platform="bigquery",
            resource_type="BigQueryLabelInfo",
            platform_instance=None,
            primary_key=self.label.primary_key(),
        )

    def platform_resource_info(self) -> BigQueryLabelInfo:
        bq_label_info = BigQueryLabelInfo(
            datahub_urn=self.datahub_urn,
            managed_by_datahub=self.managed_by_datahub,
            key=self.label.key,
            value=self.label.value,
        )
        return bq_label_info

    def platform_resource(self) -> PlatformResource:
        return PlatformResource.create(
            key=self.platform_resource_key(),
            secondary_keys=[self.datahub_urn],
            value=self.platform_resource_info(),
        )


class BigQueryPlatformResourceHelper:
    def __init__(
        self,
        bq_project: Optional[str],
        graph: Optional[DataHubGraph],
    ):
        self.bq_project = bq_project
        self.graph = graph

    platform_resource_cache: cachetools.LRUCache = cachetools.LRUCache(maxsize=500)

    def get_platform_resource(
        self, platform_resource_key: PlatformResourceKey
    ) -> Optional[PlatformResource]:
        # if graph is not available we always create a new PlatformResource
        if not self.graph:
            return None
        if self.platform_resource_cache.get(platform_resource_key.primary_key):
            return self.platform_resource_cache.get(platform_resource_key.primary_key)

        platform_resource = PlatformResource.from_datahub(
            key=platform_resource_key, graph_client=self.graph
        )
        if platform_resource:
            self.platform_resource_cache[platform_resource_key.primary_key] = (
                platform_resource
            )
            return platform_resource
        return None

    def generate_label_platform_resource(
        self,
        bigquery_label: BigQueryLabel,
        tag_urn: TagUrn,
        managed_by_datahub: bool = True,
    ) -> PlatformResource:
        new_platform_resource = BigQueryLabelPlatformResource(
            datahub_urn=tag_urn.urn(),
            project=self.bq_project,
            managed_by_datahub=managed_by_datahub,
            label=bigquery_label,
        )

        platform_resource = self.get_platform_resource(
            new_platform_resource.platform_resource_key()
        )
        if platform_resource:
            if (
                platform_resource.resource_info
                and platform_resource.resource_info.value
            ):
                try:
                    existing_info: Optional[BigQueryLabelInfo] = (
                        platform_resource.resource_info.value.as_pydantic_object(  # type: ignore
                            BigQueryLabelInfo
                        )
                    )
                except ValidationError as e:
                    logger.error(
                        f"Error converting existing value to BigQueryLabelInfo: {e}. Creating new one. Maybe this is because of a non backward compatible schema change."
                    )
                    existing_info = None

                if existing_info:
                    if (
                        new_platform_resource.platform_resource_info() == existing_info
                        or existing_info.managed_by_datahub
                    ):
                        return platform_resource
                    else:
                        raise ValueError(
                            f"Datahub URN mismatch for platform resources. Old (existing) platform resource: {platform_resource} and new platform resource: {new_platform_resource}"
                        )

        logger.info(f"Created platform resource {new_platform_resource}")

        self.platform_resource_cache.update(
            {
                new_platform_resource.platform_resource_key().primary_key: new_platform_resource.platform_resource()
            }
        )

        return new_platform_resource.platform_resource()



import json
import logging
from dataclasses import dataclass, field
from datetime import datetime
from functools import lru_cache
from typing import Any, Dict, Iterable, List, Optional

import dateutil.parser as dp
import requests
from pydantic import BaseModel
from pydantic.class_validators import root_validator, validator
from pydantic.fields import Field

from datahub.configuration.common import AllowDenyPattern
from datahub.configuration.source_common import (
    EnvConfigMixin,
    PlatformInstanceConfigMixin,
)
from datahub.emitter.mce_builder import (
    make_chart_urn,
    make_dashboard_urn,
    make_data_platform_urn,
    make_dataset_urn,
    make_dataset_urn_with_platform_instance,
    make_domain_urn,
    make_schema_field_urn,
    make_user_urn,
)
from datahub.emitter.mcp_builder import add_domain_to_entity_wu
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import MetadataWorkUnitProcessor
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.sql.sql_types import resolve_sql_type
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalHandler,
    StaleEntityRemovalSourceReport,
    StatefulStaleMetadataRemovalConfig,
)
from datahub.ingestion.source.state.stateful_ingestion_base import (
    StatefulIngestionConfigBase,
    StatefulIngestionSourceBase,
)
from datahub.metadata.com.linkedin.pegasus2avro.common import (
    ChangeAuditStamps,
    Status,
    TimeStamp,
)
from datahub.metadata.com.linkedin.pegasus2avro.metadata.snapshot import (
    ChartSnapshot,
    DashboardSnapshot,
    DatasetSnapshot,
)
from datahub.metadata.com.linkedin.pegasus2avro.mxe import MetadataChangeEvent
from datahub.metadata.com.linkedin.pegasus2avro.schema import (
    MySqlDDL,
    NullType,
    SchemaField,
    SchemaFieldDataType,
    SchemaMetadata,
)
from datahub.metadata.schema_classes import (
    AuditStampClass,
    ChartInfoClass,
    ChartTypeClass,
    DashboardInfoClass,
    DatasetLineageTypeClass,
    DatasetPropertiesClass,
    FineGrainedLineageClass,
    FineGrainedLineageDownstreamTypeClass,
    FineGrainedLineageUpstreamTypeClass,
    GlobalTagsClass,
    OwnerClass,
    OwnershipClass,
    OwnershipTypeClass,
    TagAssociationClass,
    UpstreamClass,
    UpstreamLineageClass,
)
from datahub.sql_parsing.sqlglot_lineage import (
    SqlParsingResult,
    create_lineage_sql_parsed_result,
)
from datahub.utilities import config_clean
from datahub.utilities.lossy_collections import LossyList
from datahub.utilities.registries.domain_registry import DomainRegistry

logger = logging.getLogger(__name__)

PAGE_SIZE = 25


chart_type_from_viz_type = {
    "line": ChartTypeClass.LINE,
    "big_number": ChartTypeClass.LINE,
    "table": ChartTypeClass.TABLE,
    "dist_bar": ChartTypeClass.BAR,
    "area": ChartTypeClass.AREA,
    "bar": ChartTypeClass.BAR,
    "pie": ChartTypeClass.PIE,
    "histogram": ChartTypeClass.HISTOGRAM,
    "big_number_total": ChartTypeClass.LINE,
    "dual_line": ChartTypeClass.LINE,
    "line_multi": ChartTypeClass.LINE,
    "treemap": ChartTypeClass.AREA,
    "box_plot": ChartTypeClass.BAR,
}


platform_without_databases = ["druid"]


@dataclass
class SupersetSourceReport(StaleEntityRemovalSourceReport):
    filtered: LossyList[str] = field(default_factory=LossyList)

    def report_dropped(self, name: str) -> None:
        self.filtered.append(name)


class SupersetDataset(BaseModel):
    id: int
    table_name: str
    changed_on_utc: Optional[str] = None
    explore_url: Optional[str] = ""

    @property
    def modified_dt(self) -> Optional[datetime]:
        if self.changed_on_utc:
            return dp.parse(self.changed_on_utc)
        return None

    @property
    def modified_ts(self) -> Optional[int]:
        if self.modified_dt:
            return int(self.modified_dt.timestamp() * 1000)
        return None


class SupersetConfig(
    StatefulIngestionConfigBase, EnvConfigMixin, PlatformInstanceConfigMixin
):
    # See the Superset /security/login endpoint for details
    # https://superset.apache.org/docs/rest-api
    connect_uri: str = Field(
        default="http://localhost:8088", description="Superset host URL."
    )
    display_uri: Optional[str] = Field(
        default=None,
        description="optional URL to use in links (if `connect_uri` is only for ingestion)",
    )
    domain: Dict[str, AllowDenyPattern] = Field(
        default=dict(),
        description="regex patterns for tables to filter to assign domain_key. ",
    )
    dataset_pattern: AllowDenyPattern = Field(
        default=AllowDenyPattern.allow_all(),
        description="Regex patterns for dataset to filter in ingestion.",
    )
    chart_pattern: AllowDenyPattern = Field(
        AllowDenyPattern.allow_all(),
        description="Patterns for selecting chart names that are to be included",
    )
    dashboard_pattern: AllowDenyPattern = Field(
        AllowDenyPattern.allow_all(),
        description="Patterns for selecting dashboard names that are to be included",
    )
    username: Optional[str] = Field(default=None, description="Superset username.")
    password: Optional[str] = Field(default=None, description="Superset password.")
    # Configuration for stateful ingestion
    stateful_ingestion: Optional[StatefulStaleMetadataRemovalConfig] = Field(
        default=None, description="Superset Stateful Ingestion Config."
    )
    ingest_dashboards: bool = Field(
        default=True, description="Enable to ingest dashboards."
    )
    ingest_charts: bool = Field(default=True, description="Enable to ingest charts.")
    ingest_datasets: bool = Field(
        default=False, description="Enable to ingest datasets."
    )

    provider: str = Field(default="db", description="Superset provider.")
    options: Dict = Field(default={}, description="")

    timeout: int = Field(
        default=10, description="Timeout of single API call to superset."
    )

    # TODO: Check and remove this if no longer needed.
    # Config database_alias is removed from sql sources.
    database_alias: Dict[str, str] = Field(
        default={},
        description="Can be used to change mapping for database names in superset to what you have in datahub",
    )

    class Config:
        # This is required to allow preset configs to get parsed
        extra = "allow"

    @validator("connect_uri", "display_uri")
    def remove_trailing_slash(cls, v):
        return config_clean.remove_trailing_slashes(v)

    @root_validator(skip_on_failure=True)
    def default_display_uri_to_connect_uri(cls, values):
        base = values.get("display_uri")
        if base is None:
            values["display_uri"] = values.get("connect_uri")
        return values


def get_metric_name(metric):
    if not metric:
        return ""
    if isinstance(metric, str):
        return metric
    label = metric.get("label")
    if not label:
        return ""
    return label


def get_filter_name(filter_obj):
    sql_expression = filter_obj.get("sqlExpression")
    if sql_expression:
        return sql_expression

    clause = filter_obj.get("clause")
    column = filter_obj.get("subject")
    operator = filter_obj.get("operator")
    comparator = filter_obj.get("comparator")
    return f"{clause} {column} {operator} {comparator}"


@platform_name("Superset")
@config_class(SupersetConfig)
@support_status(SupportStatus.CERTIFIED)
@capability(
    SourceCapability.DELETION_DETECTION, "Optionally enabled via stateful_ingestion"
)
@capability(SourceCapability.DOMAINS, "Enabled by `domain` config to assign domain_key")
@capability(SourceCapability.LINEAGE_COARSE, "Supported by default")
class SupersetSource(StatefulIngestionSourceBase):
    """
    This plugin extracts the following:
    - Charts, dashboards, and associated metadata

    See documentation for superset's /security/login at https://superset.apache.org/docs/rest-api for more details on superset's login api.
    """

    config: SupersetConfig
    report: SupersetSourceReport
    platform = "superset"

    def __hash__(self):
        return id(self)

    def __init__(self, ctx: PipelineContext, config: SupersetConfig):
        super().__init__(config, ctx)
        self.config = config
        self.report = SupersetSourceReport()
        if self.config.domain:
            self.domain_registry = DomainRegistry(
                cached_domains=[domain_id for domain_id in self.config.domain],
                graph=self.ctx.graph,
            )
        self.session = self.login()
        self.owner_info = self.parse_owner_info()

    def login(self) -> requests.Session:
        login_response = requests.post(
            f"{self.config.connect_uri}/api/v1/security/login",
            json={
                "username": self.config.username,
                "password": self.config.password,
                "refresh": True,
                "provider": self.config.provider,
            },
        )

        self.access_token = login_response.json()["access_token"]
        logger.debug("Got access token from superset")

        requests_session = requests.Session()
        requests_session.headers.update(
            {
                "Authorization": f"Bearer {self.access_token}",
                "Content-Type": "application/json",
                "Accept": "*/*",
            }
        )

        test_response = requests_session.get(
            f"{self.config.connect_uri}/api/v1/dashboard/",
            timeout=self.config.timeout,
        )
        if test_response.status_code != 200:
            # throw an error and terminate ingestion,
            # cannot proceed without access token
            logger.error(
                f"Failed to log in to Superset with status: {test_response.status_code}"
            )
        return requests_session

    def paginate_entity_api_results(self, entity_type, page_size=100):
        current_page = 0
        total_items = page_size

        while current_page * page_size < total_items:
            response = self.session.get(
                f"{self.config.connect_uri}/api/v1/{entity_type}",
                params={"q": f"(page:{current_page},page_size:{page_size})"},
                timeout=self.config.timeout,
            )

            if response.status_code != 200:
                logger.warning(f"Failed to get {entity_type} data: {response.text}")

            payload = response.json()
            # Update total_items with the actual count from the response
            total_items = payload.get("count", total_items)
            # Yield each item in the result, this gets passed into the construct functions
            for item in payload.get("result", []):
                yield item

            current_page += 1

    def parse_owner_info(self) -> Dict[str, Any]:
        entity_types = ["dataset", "dashboard", "chart"]
        owners_info = {}

        for entity in entity_types:
            for owner in self.paginate_entity_api_results(f"{entity}/related/owners"):
                owner_id = owner.get("value")
                if owner_id:
                    owners_info[owner_id] = owner.get("extra", {}).get("email", "")

        return owners_info

    def build_owner_urn(self, data: Dict[str, Any]) -> List[str]:
        return [
            make_user_urn(self.owner_info.get(owner.get("id"), ""))
            for owner in data.get("owners", [])
            if owner.get("id")
        ]

    @lru_cache(maxsize=None)
    def get_dataset_info(self, dataset_id: int) -> dict:
        dataset_response = self.session.get(
            f"{self.config.connect_uri}/api/v1/dataset/{dataset_id}",
            timeout=self.config.timeout,
        )
        if dataset_response.status_code != 200:
            logger.warning(f"Failed to get dataset info: {dataset_response.text}")
            return {}
        return dataset_response.json()

    def get_datasource_urn_from_id(
        self, dataset_response: dict, platform_instance: str
    ) -> str:
        schema_name = dataset_response.get("result", {}).get("schema")
        table_name = dataset_response.get("result", {}).get("table_name")
        database_id = dataset_response.get("result", {}).get("database", {}).get("id")
        database_name = (
            dataset_response.get("result", {}).get("database", {}).get("database_name")
        )
        database_name = self.config.database_alias.get(database_name, database_name)

        # Druid do not have a database concept and has a limited schema concept, but they are nonetheless reported
        # from superset. There is only one database per platform instance, and one schema named druid, so it would be
        # redundant to systemically store them both in the URN.
        if platform_instance in platform_without_databases:
            database_name = None

        if platform_instance == "druid" and schema_name == "druid":
            # Follow DataHub's druid source convention.
            schema_name = None

        # If the information about the datasource is already contained in the dataset response,
        # can just return the urn directly
        if table_name and database_id:
            return make_dataset_urn(
                platform=platform_instance,
                name=".".join(
                    name for name in [database_name, schema_name, table_name] if name
                ),
                env=self.config.env,
            )

        raise ValueError("Could not construct dataset URN")

    def construct_dashboard_from_api_data(
        self, dashboard_data: dict
    ) -> DashboardSnapshot:
        dashboard_urn = make_dashboard_urn(
            platform=self.platform,
            name=str(dashboard_data["id"]),
            platform_instance=self.config.platform_instance,
        )
        dashboard_snapshot = DashboardSnapshot(
            urn=dashboard_urn,
            aspects=[Status(removed=False)],
        )

        modified_actor = f"urn:li:corpuser:{self.owner_info.get((dashboard_data.get('changed_by') or {}).get('id', -1), 'unknown')}"
        now = datetime.now().strftime("%I:%M%p on %B %d, %Y")
        modified_ts = int(
            dp.parse(dashboard_data.get("changed_on_utc", now)).timestamp() * 1000
        )
        title = dashboard_data.get("dashboard_title", "")
        # note: the API does not currently supply created_by usernames due to a bug
        last_modified = AuditStampClass(time=modified_ts, actor=modified_actor)

        change_audit_stamps = ChangeAuditStamps(
            created=None, lastModified=last_modified
        )
        dashboard_url = f"{self.config.display_uri}{dashboard_data.get('url', '')}"

        chart_urns = []
        raw_position_data = dashboard_data.get("position_json", "{}")
        position_data = (
            json.loads(raw_position_data) if raw_position_data is not None else {}
        )
        for key, value in position_data.items():
            if not key.startswith("CHART-"):
                continue
            chart_urns.append(
                make_chart_urn(
                    platform=self.platform,
                    name=str(value.get("meta", {}).get("chartId", "unknown")),
                    platform_instance=self.config.platform_instance,
                )
            )

        # Build properties
        custom_properties = {
            "Status": str(dashboard_data.get("status")),
            "IsPublished": str(dashboard_data.get("published", False)).lower(),
            "Owners": ", ".join(
                map(
                    lambda owner: self.owner_info.get(owner.get("id", -1), "unknown"),
                    dashboard_data.get("owners", []),
                )
            ),
            "IsCertified": str(bool(dashboard_data.get("certified_by"))).lower(),
        }

        if dashboard_data.get("certified_by"):
            custom_properties["CertifiedBy"] = dashboard_data.get("certified_by", "")
            custom_properties["CertificationDetails"] = str(
                dashboard_data.get("certification_details")
            )

        # Create DashboardInfo object
        dashboard_info = DashboardInfoClass(
            description="",
            title=title,
            charts=chart_urns,
            dashboardUrl=dashboard_url,
            customProperties=custom_properties,
            lastModified=change_audit_stamps,
        )
        dashboard_snapshot.aspects.append(dashboard_info)

        dashboard_owners_list = self.build_owner_urn(dashboard_data)
        owners_info = OwnershipClass(
            owners=[
                OwnerClass(
                    owner=urn,
                    type=OwnershipTypeClass.TECHNICAL_OWNER,
                )
                for urn in (dashboard_owners_list or [])
            ],
            lastModified=last_modified,
        )
        dashboard_snapshot.aspects.append(owners_info)

        return dashboard_snapshot

    def emit_dashboard_mces(self) -> Iterable[MetadataWorkUnit]:
        for dashboard_data in self.paginate_entity_api_results("dashboard/", PAGE_SIZE):
            try:
                dashboard_id = str(dashboard_data.get("id"))
                dashboard_title = dashboard_data.get("dashboard_title", "")

                if not self.config.dashboard_pattern.allowed(dashboard_title):
                    self.report.report_dropped(
                        f"Dashboard '{dashboard_title}' (id: {dashboard_id}) filtered by dashboard_pattern"
                    )
                    continue

                dashboard_snapshot = self.construct_dashboard_from_api_data(
                    dashboard_data
                )
            except Exception as e:
                self.report.warning(
                    f"Failed to construct dashboard snapshot. Dashboard name: {dashboard_data.get('dashboard_title')}. Error: \n{e}"
                )
                continue
            # Emit the dashboard
            mce = MetadataChangeEvent(proposedSnapshot=dashboard_snapshot)
            yield MetadataWorkUnit(id=dashboard_snapshot.urn, mce=mce)
            yield from self._get_domain_wu(
                title=dashboard_title,
                entity_urn=dashboard_snapshot.urn,
            )

    def construct_chart_from_chart_data(self, chart_data: dict) -> ChartSnapshot:
        chart_urn = make_chart_urn(
            platform=self.platform,
            name=str(chart_data["id"]),
            platform_instance=self.config.platform_instance,
        )
        chart_snapshot = ChartSnapshot(
            urn=chart_urn,
            aspects=[Status(removed=False)],
        )

        modified_actor = f"urn:li:corpuser:{self.owner_info.get((chart_data.get('changed_by') or {}).get('id', -1), 'unknown')}"
        now = datetime.now().strftime("%I:%M%p on %B %d, %Y")
        modified_ts = int(
            dp.parse(chart_data.get("changed_on_utc", now)).timestamp() * 1000
        )
        title = chart_data.get("slice_name", "")

        # note: the API does not currently supply created_by usernames due to a bug
        last_modified = AuditStampClass(time=modified_ts, actor=modified_actor)

        change_audit_stamps = ChangeAuditStamps(
            created=None, lastModified=last_modified
        )

        chart_type = chart_type_from_viz_type.get(chart_data.get("viz_type", ""))
        chart_url = f"{self.config.display_uri}{chart_data.get('url', '')}"

        datasource_id = chart_data.get("datasource_id")
        if not datasource_id:
            logger.debug(
                f"chart {chart_data['id']} has no datasource_id, skipping fetching dataset info"
            )
            datasource_urn = None
        else:
            dataset_response = self.get_dataset_info(datasource_id)
            datasource_urn = self.get_datasource_urn_from_id(
                dataset_response, self.platform
            )

        params = json.loads(chart_data.get("params", "{}"))
        metrics = [
            get_metric_name(metric)
            for metric in (params.get("metrics", []) or [params.get("metric")])
        ]
        filters = [
            get_filter_name(filter_obj)
            for filter_obj in params.get("adhoc_filters", [])
        ]
        group_bys = params.get("groupby", []) or []
        if isinstance(group_bys, str):
            group_bys = [group_bys]
        # handling List[Union[str, dict]] case
        # a dict containing two keys: sqlExpression and label
        elif isinstance(group_bys, list) and len(group_bys) != 0:
            temp_group_bys = []
            for item in group_bys:
                # if the item is a custom label
                if isinstance(item, dict):
                    item_value = item.get("label", "")
                    if item_value != "":
                        temp_group_bys.append(f"{item_value}_custom_label")
                    else:
                        temp_group_bys.append(str(item))

                # if the item is a string
                elif isinstance(item, str):
                    temp_group_bys.append(item)

            group_bys = temp_group_bys

        custom_properties = {
            "Metrics": ", ".join(metrics),
            "Filters": ", ".join(filters),
            "Dimensions": ", ".join(group_bys),
        }

        chart_info = ChartInfoClass(
            type=chart_type,
            description="",
            title=title,
            chartUrl=chart_url,
            inputs=[datasource_urn] if datasource_urn else None,
            customProperties=custom_properties,
            lastModified=change_audit_stamps,
        )
        chart_snapshot.aspects.append(chart_info)

        chart_owners_list = self.build_owner_urn(chart_data)
        owners_info = OwnershipClass(
            owners=[
                OwnerClass(
                    owner=urn,
                    type=OwnershipTypeClass.TECHNICAL_OWNER,
                )
                for urn in (chart_owners_list or [])
            ],
            lastModified=last_modified,
        )
        chart_snapshot.aspects.append(owners_info)
        return chart_snapshot

    def emit_chart_mces(self) -> Iterable[MetadataWorkUnit]:
        for chart_data in self.paginate_entity_api_results("chart/", PAGE_SIZE):
            try:
                chart_id = str(chart_data.get("id"))
                chart_name = chart_data.get("slice_name", "")

                if not self.config.chart_pattern.allowed(chart_name):
                    self.report.report_dropped(
                        f"Chart '{chart_name}' (id: {chart_id}) filtered by chart_pattern"
                    )
                    continue

                # Emit a warning if charts use data from a dataset that will be filtered out
                if self.config.dataset_pattern != AllowDenyPattern.allow_all():
                    datasource_id = chart_data.get("datasource_id")
                    if datasource_id:
                        dataset_response = self.get_dataset_info(datasource_id)
                        dataset_name = dataset_response.get("result", {}).get(
                            "table_name", ""
                        )

                        if dataset_name and not self.config.dataset_pattern.allowed(
                            dataset_name
                        ):
                            self.report.warning(
                                f"Chart '{chart_name}' (id: {chart_id}) uses dataset '{dataset_name}' which is filtered by dataset_pattern"
                            )

                chart_snapshot = self.construct_chart_from_chart_data(chart_data)

                mce = MetadataChangeEvent(proposedSnapshot=chart_snapshot)
            except Exception as e:
                self.report.warning(
                    f"Failed to construct chart snapshot. Chart name: {chart_name}. Error: \n{e}"
                )
                continue
            # Emit the chart
            yield MetadataWorkUnit(id=chart_snapshot.urn, mce=mce)
            yield from self._get_domain_wu(
                title=chart_data.get("slice_name", ""),
                entity_urn=chart_snapshot.urn,
            )

    def gen_schema_fields(self, column_data: List[Dict[str, str]]) -> List[SchemaField]:
        schema_fields: List[SchemaField] = []
        for col in column_data:
            col_type = (col.get("type") or "").lower()
            data_type = resolve_sql_type(col_type)
            if data_type is None:
                data_type = NullType()

            field = SchemaField(
                fieldPath=col.get("column_name", ""),
                type=SchemaFieldDataType(data_type),
                nativeDataType="",
                description=col.get("column_name", ""),
                nullable=True,
            )
            schema_fields.append(field)
        return schema_fields

    def gen_schema_metadata(
        self,
        dataset_response: dict,
    ) -> SchemaMetadata:
        dataset_response = dataset_response.get("result", {})
        column_data = dataset_response.get("columns", [])
        schema_metadata = SchemaMetadata(
            schemaName=dataset_response.get("table_name", ""),
            platform=make_data_platform_urn(self.platform),
            version=0,
            hash="",
            platformSchema=MySqlDDL(tableSchema=""),
            fields=self.gen_schema_fields(column_data),
        )
        return schema_metadata

    def gen_dataset_urn(self, datahub_dataset_name: str) -> str:
        return make_dataset_urn_with_platform_instance(
            platform=self.platform,
            name=datahub_dataset_name,
            platform_instance=self.config.platform_instance,
            env=self.config.env,
        )

    def generate_virtual_dataset_lineage(
        self,
        parsed_query_object: SqlParsingResult,
        datasource_urn: str,
    ) -> UpstreamLineageClass:
        cll = (
            parsed_query_object.column_lineage
            if parsed_query_object.column_lineage is not None
            else []
        )

        fine_grained_lineages: List[FineGrainedLineageClass] = []

        for cll_info in cll:
            downstream = (
                [make_schema_field_urn(datasource_urn, cll_info.downstream.column)]
                if cll_info.downstream and cll_info.downstream.column
                else []
            )
            upstreams = [
                make_schema_field_urn(column_ref.table, column_ref.column)
                for column_ref in cll_info.upstreams
            ]
            fine_grained_lineages.append(
                FineGrainedLineageClass(
                    downstreamType=FineGrainedLineageDownstreamTypeClass.FIELD,
                    downstreams=downstream,
                    upstreamType=FineGrainedLineageUpstreamTypeClass.FIELD_SET,
                    upstreams=upstreams,
                )
            )

        upstream_lineage = UpstreamLineageClass(
            upstreams=[
                UpstreamClass(
                    type=DatasetLineageTypeClass.TRANSFORMED,
                    dataset=input_table_urn,
                )
                for input_table_urn in parsed_query_object.in_tables
            ],
            fineGrainedLineages=fine_grained_lineages,
        )
        return upstream_lineage

    def generate_physical_dataset_lineage(
        self,
        dataset_response: dict,
        upstream_dataset: str,
        datasource_urn: str,
    ) -> UpstreamLineageClass:
        # To generate column level lineage, we can manually decode the metadata
        # to produce the ColumnLineageInfo
        columns = dataset_response.get("result", {}).get("columns", [])
        fine_grained_lineages: List[FineGrainedLineageClass] = []

        for column in columns:
            column_name = column.get("column_name", "")
            if not column_name:
                continue

            downstream = [make_schema_field_urn(datasource_urn, column_name)]
            upstreams = [make_schema_field_urn(upstream_dataset, column_name)]
            fine_grained_lineages.append(
                FineGrainedLineageClass(
                    downstreamType=FineGrainedLineageDownstreamTypeClass.FIELD,
                    downstreams=downstream,
                    upstreamType=FineGrainedLineageUpstreamTypeClass.FIELD_SET,
                    upstreams=upstreams,
                )
            )

        upstream_lineage = UpstreamLineageClass(
            upstreams=[
                UpstreamClass(
                    type=DatasetLineageTypeClass.TRANSFORMED,
                    dataset=upstream_dataset,
                )
            ],
            fineGrainedLineages=fine_grained_lineages,
        )
        return upstream_lineage

    def construct_dataset_from_dataset_data(
        self, dataset_data: dict
    ) -> DatasetSnapshot:
        dataset_response = self.get_dataset_info(dataset_data.get("id"))
        dataset = SupersetDataset(**dataset_response["result"])

        datasource_urn = self.get_datasource_urn_from_id(
            dataset_response, self.platform
        )
        dataset_url = f"{self.config.display_uri}{dataset_response.get('result', {}).get('url', '')}"

        modified_actor = f"urn:li:corpuser:{self.owner_info.get((dataset_data.get('changed_by') or {}).get('id', -1), 'unknown')}"
        now = datetime.now().strftime("%I:%M%p on %B %d, %Y")
        modified_ts = int(
            dp.parse(dataset_data.get("changed_on_utc", now)).timestamp() * 1000
        )
        last_modified = AuditStampClass(time=modified_ts, actor=modified_actor)

        upstream_warehouse_platform = (
            dataset_response.get("result", {}).get("database", {}).get("backend")
        )
        upstream_warehouse_db_name = (
            dataset_response.get("result", {}).get("database", {}).get("database_name")
        )

        # if we have rendered sql, we always use that and defualt back to regular sql
        sql = dataset_response.get("result", {}).get(
            "rendered_sql"
        ) or dataset_response.get("result", {}).get("sql")

        # Preset has a way of naming their platforms differently than
        # how datahub names them, so map the platform name to the correct naming
        warehouse_naming = {
            "awsathena": "athena",
            "clickhousedb": "clickhouse",
            "postgresql": "postgres",
        }

        if upstream_warehouse_platform in warehouse_naming:
            upstream_warehouse_platform = warehouse_naming[upstream_warehouse_platform]

        upstream_dataset = self.get_datasource_urn_from_id(
            dataset_response, upstream_warehouse_platform
        )

        # Sometimes the field will be null instead of not existing
        if sql == "null" or not sql:
            tag_urn = f"urn:li:tag:{self.platform}:physical"
            upstream_lineage = self.generate_physical_dataset_lineage(
                dataset_response, upstream_dataset, datasource_urn
            )
        else:
            tag_urn = f"urn:li:tag:{self.platform}:virtual"
            parsed_query_object = create_lineage_sql_parsed_result(
                query=sql,
                default_db=upstream_warehouse_db_name,
                platform=upstream_warehouse_platform,
                platform_instance=None,
                env=self.config.env,
            )
            upstream_lineage = self.generate_virtual_dataset_lineage(
                parsed_query_object, datasource_urn
            )

        dataset_info = DatasetPropertiesClass(
            name=dataset.table_name,
            description="",
            externalUrl=dataset_url,
            lastModified=TimeStamp(time=modified_ts),
        )
        global_tags = GlobalTagsClass(tags=[TagAssociationClass(tag=tag_urn)])

        aspects_items: List[Any] = []
        aspects_items.extend(
            [
                self.gen_schema_metadata(dataset_response),
                dataset_info,
                upstream_lineage,
                global_tags,
            ]
        )

        dataset_snapshot = DatasetSnapshot(
            urn=datasource_urn,
            aspects=aspects_items,
        )

        dataset_owners_list = self.build_owner_urn(dataset_data)
        owners_info = OwnershipClass(
            owners=[
                OwnerClass(
                    owner=urn,
                    type=OwnershipTypeClass.TECHNICAL_OWNER,
                )
                for urn in (dataset_owners_list or [])
            ],
            lastModified=last_modified,
        )
        aspects_items.append(owners_info)

        return dataset_snapshot

    def emit_dataset_mces(self) -> Iterable[MetadataWorkUnit]:
        for dataset_data in self.paginate_entity_api_results("dataset/", PAGE_SIZE):
            try:
                dataset_name = dataset_data.get("table_name", "")

                # Check if dataset should be filtered by dataset name
                if not self.config.dataset_pattern.allowed(dataset_name):
                    self.report.report_dropped(
                        f"Dataset '{dataset_name}' filtered by dataset_pattern"
                    )
                    continue

                dataset_snapshot = self.construct_dataset_from_dataset_data(
                    dataset_data
                )
                mce = MetadataChangeEvent(proposedSnapshot=dataset_snapshot)
            except Exception as e:
                self.report.warning(
                    f"Failed to construct dataset snapshot. Dataset name: {dataset_data.get('table_name')}. Error: \n{e}"
                )
                continue
            # Emit the dataset
            yield MetadataWorkUnit(id=dataset_snapshot.urn, mce=mce)
            yield from self._get_domain_wu(
                title=dataset_data.get("table_name", ""),
                entity_urn=dataset_snapshot.urn,
            )

    def get_workunits_internal(self) -> Iterable[MetadataWorkUnit]:
        if self.config.ingest_dashboards:
            yield from self.emit_dashboard_mces()
        if self.config.ingest_charts:
            yield from self.emit_chart_mces()
        if self.config.ingest_datasets:
            yield from self.emit_dataset_mces()

    def get_workunit_processors(self) -> List[Optional[MetadataWorkUnitProcessor]]:
        return [
            *super().get_workunit_processors(),
            StaleEntityRemovalHandler.create(
                self, self.config, self.ctx
            ).workunit_processor,
        ]

    def get_report(self) -> StaleEntityRemovalSourceReport:
        return self.report

    def _get_domain_wu(self, title: str, entity_urn: str) -> Iterable[MetadataWorkUnit]:
        domain_urn = None
        for domain, pattern in self.config.domain.items():
            if pattern.allowed(title):
                domain_urn = make_domain_urn(
                    self.domain_registry.get_domain_urn(domain)
                )
                break

        if domain_urn:
            yield from add_domain_to_entity_wu(
                entity_urn=entity_urn,
                domain_urn=domain_urn,
            )



import json
import logging
from typing import Optional

import click

from datahub.configuration.common import ConfigModel
from datahub.configuration.config_loader import load_config_file
from datahub.emitter.mce_builder import datahub_guid

logger = logging.getLogger(__name__)


def _make_ingestion_urn(name: str) -> str:
    guid = datahub_guid(
        {
            "name": name,
        }
    )
    return f"urn:li:dataHubIngestionSource:deploy-{guid}"


class DeployOptions(ConfigModel):
    name: str
    schedule: Optional[str] = None
    time_zone: str = "UTC"
    cli_version: Optional[str] = None
    executor_id: str = "default"


def deploy_source_vars(
    name: Optional[str],
    config: str,
    urn: Optional[str],
    executor_id: str,
    cli_version: Optional[str],
    schedule: Optional[str],
    time_zone: str,
    extra_pip: Optional[str],
    debug: bool = False,
) -> dict:
    pipeline_config = load_config_file(
        config,
        allow_stdin=True,
        allow_remote=True,
        resolve_env_vars=False,
    )

    deploy_options_raw = pipeline_config.pop("deployment", None)
    if deploy_options_raw is not None:
        deploy_options = DeployOptions.parse_obj(deploy_options_raw)

        if name:
            logger.info(f"Overriding deployment name {deploy_options.name} with {name}")
            deploy_options.name = name
    else:
        if not name:
            raise click.UsageError(
                "Either --name must be set or deployment_name specified in the config"
            )
        deploy_options = DeployOptions(name=name)

    # Use remaining CLI args to override deploy_options
    if schedule:
        deploy_options.schedule = schedule
    if time_zone:
        deploy_options.time_zone = time_zone
    if cli_version:
        deploy_options.cli_version = cli_version
    if executor_id:
        deploy_options.executor_id = executor_id

    logger.info(f"Using {repr(deploy_options)}")

    if not urn:
        # When urn/name is not specified, we will generate a unique urn based on the deployment name.
        urn = _make_ingestion_urn(deploy_options.name)
        logger.info(f"Using recipe urn: {urn}")

    variables: dict = {
        "urn": urn,
        "input": {
            "name": deploy_options.name,
            "type": pipeline_config["source"]["type"],
            "config": {
                "recipe": json.dumps(pipeline_config),
                "executorId": deploy_options.executor_id,
                "debugMode": debug,
                "version": deploy_options.cli_version,
            },
        },
    }

    if deploy_options.schedule is not None:
        variables["input"]["schedule"] = {
            "interval": deploy_options.schedule,
            "timezone": deploy_options.time_zone,
        }
    if extra_pip is not None:
        extra_args_list = (
            variables.get("input", {}).get("config", {}).get("extraArgs", [])
        )
        extra_args_list.append({"key": "extra_pip_requirements", "value": extra_pip})
        variables["input"]["config"]["extraArgs"] = extra_args_list

    return variables



from functools import singledispatchmethod

from datahub.api.entities.assertion.assertion_operator import (
    BetweenOperator,
    EqualToOperator,
    GreaterThanOperator,
    GreaterThanOrEqualToOperator,
    IsFalseOperator,
    IsNullOperator,
    IsTrueOperator,
    LessThanOperator,
    LessThanOrEqualToOperator,
    NotNullOperator,
    Operators,
)


class SnowflakeMetricEvalOperatorSQLGenerator:
    @singledispatchmethod
    def operator_sql(self, operators: Operators, metric_sql: str) -> str:
        """
        Generates Operator SQL that applies operator on `metric`
        and returns a numeric boolean value 1 if PASS, 0 if FAIL

        """
        raise ValueError(f"Unsupported metric operator type {type(operators)} ")

    @operator_sql.register
    def _(self, operators: EqualToOperator, metric_sql: str) -> str:
        return f"select case when metric={operators.value} then 1 else 0 end from ({metric_sql})"

    @operator_sql.register
    def _(self, operators: BetweenOperator, metric_sql: str) -> str:
        return f"select case when metric between {operators.min} and {operators.max} then 1 else 0 end from ({metric_sql})"

    @operator_sql.register
    def _(self, operators: LessThanOperator, metric_sql: str) -> str:
        return f"select case when metric < {operators.value} then 1 else 0 end from ({metric_sql})"

    @operator_sql.register
    def _(self, operators: LessThanOrEqualToOperator, metric_sql: str) -> str:
        return f"select case when metric <= {operators.value} then 1 else 0 end from ({metric_sql})"

    @operator_sql.register
    def _(self, operators: GreaterThanOperator, metric_sql: str) -> str:
        return f"select case when metric > {operators.value} then 1 else 0 end from ({metric_sql})"

    @operator_sql.register
    def _(self, operators: GreaterThanOrEqualToOperator, metric_sql: str) -> str:
        return f"select case when metric >= {operators.value} then 1 else 0 end from ({metric_sql})"

    @operator_sql.register
    def _(self, operators: NotNullOperator, metric_sql: str) -> str:
        return (
            f"select case when metric is not null then 1 else 0 end from ({metric_sql})"
        )

    @operator_sql.register
    def _(self, operators: IsNullOperator, metric_sql: str) -> str:
        return f"select case when metric is null then 1 else 0 end from ({metric_sql})"

    @operator_sql.register
    def _(self, operators: IsTrueOperator, metric_sql: str) -> str:
        return f"select case when metric then 1 else 0 end from ({metric_sql})"

    @operator_sql.register
    def _(self, operators: IsFalseOperator, metric_sql: str) -> str:
        return f"select case when not metric then 1 else 0 end from ({metric_sql})"
